# inserting milestones of a fixed len () into the texts
import re
import os
import math
import sys
from itertools import groupby
import openiti.helper.ara as ara


def milestones(file, length, last_ms_cnt):
    # ara_regex = re.compile("^[ذ١٢٣٤٥٦٧٨٩٠ّـضصثقفغعهخحجدًٌَُلإإشسيبلاتنمكطٍِلأأـئءؤرلاىةوزظْلآآ]+$")
    # new char list without ZERO WIDTH NON-JOINER and ZERO WIDTH JOINER
    # ara_regex = re.compile("[ءآأؤإئابةتثجحخدذرزسشصضطظعغـفقكلمنهوىيًٌٍَُِّْ٠١٢٣٤٥٦٧٨٩ٮٰٹپچژکگیے۱۲۳۴۵]+")
    splitter = "#META#Header#End#"
    file_name = re.split("-[a-z]{3}\d{1}(\.(mARkdown|inProgress|completed))?$", file.split("/")[-1])[0]
    print(file_name)
    if re.search("[A-Z]{1}$", file_name):
        continuous = True
    else:
        continuous = False

    with open(file, "r", encoding="utf8") as f:
        data = f.read()

        # splitter test
        if splitter in data:
            data_parts = re.split("\n*#META#Header#End#\n*", data)
            head = data_parts[0]
            # remove the final new line and spaces to avoid having the milestone tag in a new empty line
            text = data_parts[1].rstrip()
            # remove old milestone ids. Fixed strings, until we make them as user input, if required!
            text = re.sub(" Milestone300", "", text)
            text = re.sub(" ms[A-Z]?\d+", "", text)

            # insert Milestones
            ara_toks_count = ara.ar_tok_cnt(text)
            # ara_toks_count = len(ara_regex.findall(text))

            ms_tag_str_len = len(str(math.floor(ara_toks_count / length)))
            # find all tokens to check the Arabic tokens in their positions
            all_toks = re.findall(r"\w+|\W+", text)

            token_count = 0
            ms_count = last_ms_cnt

            new_data = []
            for i in range(0, len(all_toks)):
                # check each token at its position
                # if ara.ar_tok.search(all_toks[i]):
                # if ara_regex.search(all_toks[i]):
                if re.search(ara.ar_tok, all_toks[i]):
                    token_count += 1
                    new_data.append(all_toks[i])
                else:
                    new_data.append(all_toks[i])

                if token_count == length or i == len(all_toks) - 1:
                    ms_count += 1
                    if continuous:
                        milestone = " ms" + file_name[-1] + str(ms_count).zfill(ms_tag_str_len)
                    else:
                        milestone = " ms" + str(ms_count).zfill(ms_tag_str_len)
                    new_data.append(milestone)
                    token_count = 0

            ms_text = "".join(new_data)

            test = re.sub(" ms[A-Z]?\d+", "", ms_text)
            if test == text:
                # print("\t\tThe file has not been damaged!")
                # Milestones TEST
                ms = re.findall("ms[A-Z]?\d+", ms_text)
                print("\t\t%d milestones (%d words)" % (len(ms), length))
                ms_text = head.rstrip() + "\n\n" + splitter + "\n\n" + ms_text
                with open(file, "w", encoding="utf8") as f9:
                    f9.write(ms_text)
                return ms_count
            else:
                print("\t\tSomething got messed up...")
                return -1

        else:
            print("The file is missing the splitter!")
            print(file)
            return -1


def groupby_books(name):
    b_id = re.split("-[a-z]{3}\d{1}", name)[0]
    if re.search("[A-Z]{1}$", b_id):
        return b_id[:-1]
    else:
        return b_id


def process_files(main_folder, ms_len):
# main_folder = sys.argv[1]
#     ms_len = 300

    # The pattern is to remove the old ms ids. For now, we have it fixed, so no need to pass it as an argument to
    # milestone func. For any changes in the future, this can be a user input to decide what to remove from and
    # insert into
    # the text.
    # ms_pattern = " ms[A-Z]?\d+"

    if not os.path.exists(main_folder):
            print("invalid path: ", main_folder)
            sys.exit(1)
    else:
        # process all texts in OpenITI
        for root, dirs, files in os.walk(main_folder):
            book_files = [f for f in files if
                          re.search("^\d{4}\w+\.\w+\.\w+-[a-z]{3}\d{1}(\.(mARkdown|inProgress|completed))?$", f)]
            # books = map(lambda x: re.split("-[a-z]{3}\d{1}", x)[0], book_files)
            grouped_books = [list(items) for gr, items in groupby(sorted(book_files), key=lambda name: groupby_books(name))]

            for group in grouped_books:
                if not all(re.search("[A-Z]$", re.split("-[a-z]{3}\d{1}", x)[0]) for x in sorted(group)):
                    for g in group:
                        milestones(os.path.join(root, g), ms_len, 0)

                elif any(re.search("[A-Z]$", re.split("-[a-z]{3}\d{1}", x)[0]) for x in sorted(group)):
                    # prev_ms_cnt = 0
                    group.sort(key=lambda f: f.split("-")[1])
                    grouped_extensions = [list(items) for gr, items in groupby(group,
                                                                               key=lambda name: name.split("-")[1])]
                    for sub_g in grouped_extensions:
                        prev_ms_cnt = 0
                        for f in sorted(sub_g):
                            prev_ms_cnt = milestones(os.path.join(root, f), ms_len, prev_ms_cnt)
    # return main_folder


if __name__ == '__main__':
    folder = input("Enter the path to the OpenITI folder: ")
    ms_length = input("Enter the length of milestones: ")
    # TODO:
    #  skip for the time being as we are not sure whether some Arabic content is being changed since last run
    # re_insert = input("Do you want to re-insert milestone ids to a file with milestone ids? (Y or N)")
    process_files(folder, int(ms_length))
    print("Done!")




"""Classes and functions to work with OpenITI URIs


Todo:
    * make the print output look nicer
    * reflow texts + insert milestones?
    * compare with Maxim's script: _add_new_text_from_folder.py
      in maintenance repo


The Module contains a URI class that represents an OpenITI URI as an object.
The URI class's methods allow

* Checking whether all components of the URI are valid
* Accessing and changing components of the URI
* Getting the URI's current uri_type ("author", "book", "version", None)
* Building different versions of the URI ("author", "book", etc.)
* Building paths based on the URI

In addition to the URI class, the module contains a number of functions for
implementing URI changes in the OpenITI corpus



Examples:
    The Module contains a URI class that represents an OpenITI URI as an object.
    By calling the URI class, you create an instance of the URI class

    >>> from uri import URI
    >>> instance1 = URI("0255Jahiz.Bayan")
    >>> instance2 = URI("0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed")

    The URI instance inherits all the URI class's methods and properties:

    - Making string representations of a URI instance: print() and repr():

    >>> t = URI("0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed")
    >>> print(repr(t))
    uri(date:0255, author:Jahiz, title:Hayawan, version:Sham19Y0023775, language:ara, edition_no:1, extension:completed)
    >>> print(t)
    0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed

    - Accessing components of the URI:

    >>> t.author
    'Jahiz'
    >>> t.date
    '0255'

    - Changing components of the URI by setting properties to a new value:

    >>> u = URI("0255Jahiz.Hayawan")
    >>> u.author = "JahizBasri"
    >>> print(u)
    0255JahizBasri.Hayawan

    - Validity tests: setting invalid values for part of a uri returns an error\
    (implemented for instantiation of URI objects + setting URI components)::

        # >>> URI("255Jahiz")
        Exception: Date Error: URI must start with a date of 4 digits (255 has 3!)

        # >>> URI("0255Jāḥiẓ")
        Exception: Author name Error: Author name (Jāḥiẓ) should not contain
        digits or non-ASCII characters(culprits: ['ā', 'ḥ', 'ẓ'])

        # >>> t.author = "Jāḥiẓ"
        Exception: Author name Error: Author name (Jāḥiẓ) should not contain
        digits or non-ASCII characters(culprits: ['ā', 'ḥ', 'ẓ'])

        # >>> t.author = "0255Jahiz"
        Exception: Author name Error: Author name (0255Jahiz) should not contain
        digits or non-ASCII characters(culprits: ['0', '2', '5', '5'])

        # >>> URI("0255Jahiz.Al-Hayawan")
        Exception: Book title Error: Book title (Al-Hayawan) should not contain
        non-ASCII characters(culprits: ['-'])

        # >>> URI("0255Jahiz.Hayawan.Shāmila00123545-ara1")
        Exception: Version string Error: Version string (Shāmila00123545)
        should not contain non-ASCII characters(culprits: ['ā'])

        # >>> URI("0255Jahiz.Hayawan.Shamela00123545-arab1")
        Exception: Language code (arab) should be an ISO 639-2 language code,
        consisting of 3 characters

        # >>> t.extension = "markdown"
        Exception: Extension (markdown) is not among the allowed extensions
        (['inProgress', 'completed', 'mARkdown', 'yml', ''])

        # >>> URI("0255Jahiz.Hayawan.Shamela00123545-ara1.markdown")
        Exception: Extension (markdown) is not among the allowed extensions
        (['inProgress', 'completed', 'mARkdown', 'yml', ''])

    - Getting the URI's current uri_type ("author", "book", "version", None),\
    i.e., the longest URI that can be built from the object's components:

    >>> t.uri_type
    'version'
    >>> t.language = ""
    >>> t.uri_type
    'book'
    >>> t.date = ""
    >>> t.uri_type == None
    True

    - Building different versions of the URI\
    (uri_types: "author", "author_yml", "book", "book_yml",\
    "version", "version_yml", "version_file"):

    >>> t = URI("0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed")
    >>> t.build_uri(uri_type="author")
    '0255Jahiz'
    >>> t.build_uri("book")
    '0255Jahiz.Hayawan'
    >>> t.build_uri("book_yml")
    '0255Jahiz.Hayawan.yml'
    >>> t.build_uri("version")
    '0255Jahiz.Hayawan.Sham19Y0023775-ara1'
    >>> t.build_uri("version_file")
    '0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed'

    Simply calling the URI object (i.e., writing parentheses after
    the variable name works as an alias for the build_uri function:

    >>> t = URI("0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed")
    >>> t(uri_type="author")
    '0255Jahiz'
    >>> t("book")
    '0255Jahiz.Hayawan'
    >>> t("book_yml")
    '0255Jahiz.Hayawan.yml'
    >>> t("version")
    '0255Jahiz.Hayawan.Sham19Y0023775-ara1'
    >>> t("version_file")
    '0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed'


    - Building paths based on the URI:

    >>> t.build_pth(uri_type="version", base_pth="D:\\test")
    'D:/test/0275AH/data/0255Jahiz/0255Jahiz.Hayawan'
    >>> t.build_pth(uri_type="version_file", base_pth="D:\\test")
    'D:/test/0275AH/data/0255Jahiz/0255Jahiz.Hayawan/0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed'
    >>> t.build_pth("version")
    './0275AH/data/0255Jahiz/0255Jahiz.Hayawan'
    >>> t.build_pth(uri_type="book_yml")
    './0275AH/data/0255Jahiz/0255Jahiz.Hayawan/0255Jahiz.Hayawan.yml'

    Without uri_type argument, build_pth() builds the fullest path it can:

    >>> t.build_pth()
    './0275AH/data/0255Jahiz/0255Jahiz.Hayawan/0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed'
    >>> t.language=""  # (removing the language property makes it impossible to build a version uri)
    >>> t.build_pth()
    './0275AH/data/0255Jahiz/0255Jahiz.Hayawan'

    NB: by default, build_pth() takes the OpenITI folder structure into account,
    in which authors are grouped in 25-year batches by their death date.
    If you do not want to use this feature, set the URI class's
    data_in_25_year_repos attribute to False:

    >>> t.build_pth()
    './0275AH/data/0255Jahiz/0255Jahiz.Hayawan'
    >>> t.data_in_25_year_repos = False
    >>> t.build_pth()
    './0255Jahiz/0255Jahiz.Hayawan'
    >>> t.data_in_25_year_repos = True
    
    >>> URI.data_in_25_year_repos = False
    >>> u = URI("0255Jahiz.Hayawan")
    >>> u.build_pth()
    './0255Jahiz/0255Jahiz.Hayawan'
    >>> URI.data_in_25_year_repos = True
    >>> u.build_pth()
    './0275AH/data/0255Jahiz/0255Jahiz.Hayawan'

    ---------------------------------------------------------------------------


    In addition to the URI class, the module contains a function for    
    implementing URI changes in the OpenITI corpus: change_uri.
    
    The function has an *execute* flag. If set to False,\
    the function will not immediately be executed but first show all changes\
    it will make, and then ask the user whether to carry out the changes or not.
    
    If a version URI changes:
    
        * new author and book folders are made if necessary
          (including the creation of new author and book yml files)
        * all text files related that version should be moved
        * the yml file of that version should be updated and moved
        
    If a book uri changes:
    
        * new author and book folders are made if necessary
        * the yml file of that book should be updated and moved
        * all text files of all versions of the book should be moved
        * all yml files of versions of that book should be updated and moved
        * the original book folder itself should be (re)moved
        
    if an author uri changes:

        * new author and book folders are made if necessary
        * the yml file of that author should be updated and moved
        * all book yml files of that should be updated and moved
        * all annotation text files of all versions of all books should be moved
        * all yml files of versions of all books should be updated and moved
        * the original book folders should be (re)moved
        * the original author folder itself should be (re)moved

    Examples::
    
        change_uri("0255Jahiz", "0256Jahiz")
        change_uri("0255Jahiz", "0255JahizBasri")
        change_uri("0255Jahiz.Hayawan", "0255Jahiz.KitabHayawan")
        change_uri("0255Jahiz.Hayawan.Shamela002526-ara1",
                   "0255Jahiz.Hayawan.Shamela002526-ara2")
        change_uri("0255Jahiz.Hayawan.Shamela002526-ara1.completed",
                   "0255Jahiz.Hayawan.Shamela002526-ara1.mARkdown")
"""

import copy
import os
import re
import shutil

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.dirname(path.abspath(__file__))))
    sys.path.append(root_folder)

from openiti.helper.funcs import read_header, get_all_yml_files_in_folder, get_all_text_files_in_folder
from openiti.helper.ara import ar_cnt_file
from openiti.helper.templates import author_yml_template, book_yml_template, \
                                     version_yml_template, readme_template, \
                                     text_questionnaire_template
from openiti.helper import yml


os.sep = "/"
ISO_CODES = re.split("[\n\r\s]+",
                     """aar abk ace ach ada ady afa afh afr ain aka akk alb sqi
ale alg alt amh ang anp apa ara arc arg arm hye arn arp art arw asm ast ath aus
ava ave awa aym aze bad bai bak bal bam ban baq eus bas bat bej bel bem ben ber
bho bih bik bin bis bla bnt tib bod bos bra bre btk bua bug bul bur mya byn cad
cai car cat cau ceb cel cze ces cha chb che chg chi zho chk chm chn cho chp chr
chu chv chy cmc cnr cop cor cos cpe cpf cpp cre crh crp csb cus wel cym cze ces
dak dan dar day del den ger deu dgr din div doi dra dsb dua dum dut nld dyu dzo
efi egy eka gre ell elx eng enm epo est baq eus ewe ewo fan fao per fas fat fij
fil fin fiu fon fre fra fre fra frm fro frr frs fry ful fur gaa gay gba gem geo
kat ger deu gez gil gla gle glg glv gmh goh gon gor got grb grc gre ell grn gsw
guj gwi hai hat hau haw heb her hil him hin hit hmn hmo hrv hsb hun hup arm hye
iba ibo ice isl ido iii ijo iku ile ilo ina inc ind ine inh ipk ira iro ice isl
ita jav jbo jpn jpr jrb kaa kab kac kal kam kan kar kas geo kat kau kaw kaz kbd
kha khi khm kho kik kin kir kmb kok kom kon kor kos kpe krc krl kro kru kua kum
kur kut lad lah lam lao lat lav lez lim lin lit lol loz ltz lua lub lug lui lun
luo lus mac mkd mad mag mah mai mak mal man mao mri map mar mas may msa mdf mdr
men mga mic min mis mac mkd mkh mlg mlt mnc mni mno moh mon mos mao mri may msa
mul mun mus mwl mwr bur mya myn myv nah nai nap nau nav nbl nde ndo nds nep new
nia nic niu dut nld nno nob nog non nor nqo nso nub nwc nya nym nyn nyo nzi oci
oji ori orm osa oss ota oto paa pag pal pam pan pap pau peo per fas phi phn pli
pol pon por pra pro pus qaa qtz que raj rap rar roa roh rom rum ron rum ron run
rup rus sad sag sah sai sal sam san sas sat scn sco sel sem sga sgn shn sid sin
sio sit sla slo slk slo slk slv sma sme smi smj smn smo sms sna snd snk sog som
son sot spa alb sqi srd srn srp srr ssa ssw suk sun sus sux swa swe syc syr tah
tai tam tat tel tem ter tet tgk tgl tha tib bod tig tir tiv tkl tlh tli tmh tog
ton tpi tsi tsn tso tuk tum tup tur tut tvl twi tyv udm uga uig ukr umb und urd
uzb vai ven vie vol vot wak wal war was wel cym wen wln wol xal xho yao yap yid
yor ypk zap zbl zen zgh zha chi zho znd zul zun zxx zza""")

extensions = ["inProgress", "completed", "mARkdown", "yml", "",
              "pdf", "zip", "rar"]

created_folders = []
created_ymls = []

class URI:
    """
    A class that represents the OpenITI URI as a Python object.

    OpenITI URIs consist of the following elements:
    0768IbnMuhammadTaqiDinBaclabakki.Hadith.Shamela0009426-ara1.mARkdown
    
        * VersionURI: consists of
        
          - EditionURI: consists of
          
            * Work URI: consists of
            
              - AuthorID: consists of
              
                * author's death date (self.date): 0768
                * shuhra of the author (self.author): IbnMuhammadTaqiDinBaclabakki
                
              - BookID (self.title): Hadith: short title of the book
              
            * Version URI: consists of
            
              - VersionID (self.version): Shamela0009426: ID of the collection/contributor
                from which we got the book + number of the book in that collection
                
              - Lang:
              
                * self.language: ara: ISO 639-2 language code
                * self.edition_no: 1: edition version number
                  (different digitizations of the same edition get the same edition_no)
                  
        * self.extension = mARkdown (can be inProgress, mARkdown, completed, "")

    Examples:
        >>> from uri import URI
        >>> t = URI("0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed")

        Representations of a URI object: print() and repr():

        >>> print(repr(t))
        uri(date:0255, author:Jahiz, title:Hayawan, version:Sham19Y0023775, language:ara, edition_no:1, extension:completed)
        >>> print(t)
        0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed

        Accessing components of the URI:
        
        >>> t.author
        'Jahiz'
        >>> t.date
        '0255'

        Getting URI's current uri_type ("author", "book", "version", None),
        i.e., the longest URI that can be built from the object's components:

        >>> t.uri_type
        'version'
        >>> t.language = ""
        >>> t.uri_type
        'book'
        >>> t.date = ""
        >>> t.uri_type == None
        True

        Building different versions of the URI
        (uri_types: "author", "author_yml", "book", "book_yml",
        "version", "version_yml", "version_file"):

        >>> t = URI("0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed")
        >>> t.build_uri(uri_type="author")
        '0255Jahiz'
        >>> t.build_uri("book")
        '0255Jahiz.Hayawan'
        >>> t.build_uri("book_yml")
        '0255Jahiz.Hayawan.yml'
        >>> t.build_uri("version")
        '0255Jahiz.Hayawan.Sham19Y0023775-ara1'
        >>> t.build_uri("version_file")
        '0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed'

        Building paths based on the URI:

        >>> t.build_pth(uri_type="version", base_pth="D:\\test")
        'D:/test/0275AH/data/0255Jahiz/0255Jahiz.Hayawan'
        >>> t.build_pth(uri_type="version_file", base_pth="D:\\test")
        'D:/test/0275AH/data/0255Jahiz/0255Jahiz.Hayawan/0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed'
        >>> t.build_pth("version")
        './0275AH/data/0255Jahiz/0255Jahiz.Hayawan'
        >>> t.build_pth(uri_type="book_yml")
        './0275AH/data/0255Jahiz/0255Jahiz.Hayawan/0255Jahiz.Hayawan.yml'

        Without uri_type argument, build_pth() builds the fullest path it can:

        >>> t.build_pth()
        './0275AH/data/0255Jahiz/0255Jahiz.Hayawan/0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed'
        >>> t.language=""  # (removing the language property makes it impossible to build a version uri)
        >>> t.build_pth()
        './0275AH/data/0255Jahiz/0255Jahiz.Hayawan'

        NB: by default, build_pth() takes the OpenITI folder structure
        into account, in which authors are grouped in 25-year batches
        by their death date.
        If you do not want to use this feature, set the URI class's
        data_in_25_year_repos attribute to False:

        >>> t.build_pth()
        './0275AH/data/0255Jahiz/0255Jahiz.Hayawan'
        >>> t.data_in_25_year_repos = False
        >>> t.build_pth()
        './0255Jahiz/0255Jahiz.Hayawan'
        >>> t.language="ara"
        >>> t.build_pth()
        './0255Jahiz/0255Jahiz.Hayawan'
        >>> t.data_in_25_year_repos = True
        
        >>> URI.data_in_25_year_repos = False
        >>> u = URI("0255Jahiz.Hayawan")
        >>> u.build_pth()
        './0255Jahiz/0255Jahiz.Hayawan'
        >>> URI.data_in_25_year_repos = True
        >>> u.build_pth()
        './0275AH/data/0255Jahiz/0255Jahiz.Hayawan'

        Validity tests: setting invalid values for part of a uri returns an error::

            # >>> URI("255Jahiz")
            Exception: Date Error: URI must start with a date of 4 digits (255 has 3!)
            
            # >>> URI("0255Jāḥiẓ")
            Exception: Author name Error: Author name (Jāḥiẓ) should not contain
            digits or non-ASCII characters(culprits: ['ā', 'ḥ', 'ẓ'])
            
            # >>> t.author = "Jāḥiẓ"
            Exception: Author name Error: Author name (Jāḥiẓ) should not contain
            digits or non-ASCII characters(culprits: ['ā', 'ḥ', 'ẓ'])
            
            # >>> t.author = "0255Jahiz"
            Exception: Author name Error: Author name (0255Jahiz) should not contain
            digits or non-ASCII characters(culprits: ['0', '2', '5', '5'])
            
            # >>> URI("0255Jahiz.Al-Hayawan")
            Exception: Book title Error: Book title (Al-Hayawan) should not contain
            non-ASCII characters(culprits: ['-'])
            
            # >>> URI("0255Jahiz.Hayawan.Shāmila00123545-ara1")
            Exception: Version string Error: Version string (Shāmila00123545)
            should not contain non-ASCII characters(culprits: ['ā'])

            # >>> URI("0255Jahiz.Hayawan.Shamela00123545-arab1")
            Exception: Language code (arab) should be an ISO 639-2 language code,
            consisting of 3 characters

            # >>> t.extension = "markdown"
            Exception: Extension (markdown) is not among the allowed extensions
            (['inProgress', 'completed', 'mARkdown', 'yml', ''])

            # >>> URI("0255Jahiz.Hayawan.Shamela00123545-ara1.markdown")
            Exception: Extension (markdown) is not among the allowed extensions
            (['inProgress', 'completed', 'mARkdown', 'yml', ''])
    """

    def __init__(self, uri_string=None):
        """Initialize the URI object and its components: if a uri_string is provided,
        it will be split into its components.

        Args:
            uri_string (str): (a path to) an OpenITI URI, e.g.,
                0768IbnMuhammadTaqiDinBaclabakki.Hadith.Shamela0009426-ara1; 
                0768IbnMuhammadTaqiDinBaclabakki.Hadith; 
                0768IbnMuhammadTaqiDinBaclabakki; 
                D:\OpenITI\25Yrepos\data\0275AH\0255Jahiz. 
                Defaults to None

        Examples:
            >>> uri1 = URI("0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed")
            >>> uri2 = URI("0255Jahiz.Hayawan")
            >>> uri3 = URI()
            >>> uri4 = URI(r"D:\\OpenITI\\25Yrepos\\data\\0275AH\\0255Jahiz\\0255Jahiz.Hayawan\\0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed")
            >>> print(uri4)
            0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed
            >>> print(uri4.base_pth)
            D:/OpenITI/25Yrepos/data
        """
        try:
            self.data_in_25_year_repos
        except:
            self.data_in_25_year_repos = True
        #print("init: self.data_in_25_year_repos", self.data_in_25_year_repos)
        self.date = ""
        self.author = ""
        self.title = ""
        self.version = ""
        self.language = ""
        self.edition_no = ""
        self.extension = ""
        if uri_string:
            if len(re.split(r"[\\/]", uri_string)) > 1: # deal with paths:
                self.base_pth, self.uri_string = os.path.split(uri_string)
                if self.data_in_25_year_repos:
                    # set self.base_pth to the parent of the 25Y folder:
                    if re.search("\d{4}AH", self.base_pth):
                        while not re.search("\d{4}AH",
                                            os.path.split(self.base_pth)[1]):
                            self.base_pth = os.path.split(self.base_pth)[0]
                        self.base_pth = os.path.split(self.base_pth)[0]
                    else: # if there is no 25Y folder:
                        self.base_pth, self.uri_string = os.path.split(uri_string)
                else:
                    self.base_pth, self.uri_string = os.path.split(uri_string)
                    #print("init: establishing self.base_pth")
                    #print("  ", self.base_pth)
                    #print("  split:", os.path.split(self.base_pth))
                    while re.search("\d{4}[A-Za-z]",
                                    os.path.split(self.base_pth)[1]):
                        self.base_pth = os.path.split(self.base_pth)[0]
                        #print("   >", self.base_pth)
            else:
                self.uri_string = uri_string
            self.split_uri(self.uri_string)
        else:
            self.uri_string = ""
        # make it possible to set these values for every instance of the class:
        try:
            self.base_pth
        except: # if the base_pth has not been set before instantiating the class:
            self.base_pth = "."
        #print("init: self.base_pth", self.base_pth)



    ############################################################################
    # Setter and getter methods: intercept mistakes when setting URI properties:

    @property
    def date(self):
        """Get the URI's date property"""
        return self.__date

    @date.setter
    def date(self, date):
        """Set the URI's date property, after checking its conformity."""
        self.__date = self.check_date(date)

    def check_date(self, date):
        """Check if date is valid (i.e., 4-digit number or empty string)"""
        if date == "":
            return ""
        date = str(date)
        if len(date) != 4:
            msg = "Date Error: URI must start with a date of 4 digits "
            msg += "({} has {}!)".format(date, len(date))
            raise Exception(msg)
        return date


    @property
    def author(self):
        """Get the URI's author property"""
        return self.__author

    @author.setter
    def author(self, author):
        """Set the URI's author property, after checking its conformity."""
        self.__author = self.check_ASCII_letters(author, "Author name")

    def check_ASCII_letters(self, test_string, string_type):
        """Check whether the test_string only contains ASCII letters."""
        if re.findall("[^A-Za-z]", test_string):
            msg = "{0} Error: {0} ({1}) ".format(string_type, test_string)
            msg += "should not contain digits or non-ASCII characters"
            msg += "(culprits: {})".format(re.findall("[^A-Za-z]", test_string))
            raise Exception(msg)
        return test_string


    @property
    def title(self):
        """Get the URI's title property."""
        return self.__title

    @title.setter
    def title(self, title):
        """Set the URI's title property, after checking its conformity."""
        #self.__title = self.check_ASCII_letters(title, "Book title")
        self.__title = self.check_ASCII(title, "Book title")


    @property
    def version(self):
        """Set the URI's version property."""
        return self.__version

    @version.setter
    def version(self, version):
        """Set the URI's version property, after checking its conformity."""
        self.__version = self.check_ASCII(version, "Version string")

    def check_ASCII(self, test_string, string_type):
        """Check whether the test_string only contains ASCII letters and digits."""
        if re.findall("[^A-Za-z0-9]", test_string):
            msg = "{0} Error: {0} ({1}) ".format(string_type, test_string)
            msg += "should not contain non-ASCII characters"
            msg += "(culprits: {})".format(re.findall("[^A-Za-z0-9]", test_string))
            raise Exception(msg)
        return test_string


    @property
    def language(self):
        """Get the URI's language property (an ISO 639-2 language code)."""
        return self.__language

    @language.setter
    def language(self, language):
        """Set the URI's language property, after checking its conformity."""
        self.__language = self.check_language_code(language)

    def check_language_code(self, language):
        """Check whether language is a valid ISO 639-2 language code."""
        if language == "":
            return ""
        if not language in ISO_CODES:
            msg = "Language code ({}) ".format(language)
            msg += "should be an ISO 639-2 language code, consisting of 3 characters"
            raise Exception(msg)
        return language


    @property
    def edition_no(self):
        """Get the URI's edition_no property (i.e., the last digit of the URI)"""
        return self.__edition_no

    @edition_no.setter
    def edition_no(self, edition_no):
        """Set the URI's edition_no property, after checking its conformity."""
        self.__edition_no = edition_no


    @property
    def extension(self):
        """Get the URI's extension property."""
        return self.__extension

    @extension.setter
    def extension(self, extension):
        """Set the URI's extension property, after checking its conformity."""
        self.__extension = self.check_extension(extension)

    def check_extension(self, extension):
        """Check whether the proposed extension is allowed."""
        if extension not in extensions:
            msg = "Extension ({}) ".format(extension)
            msg += "is not among the allowed extensions ({})".format(extensions)
            raise Exception(msg)
        return extension


    @property
    def base_pth(self):
        """Get the URI's base_pth property (the path to be prepended to the URIs)
        (usually the folder in which the OpenITI 25-years repos reside)"""
        return self.__base_pth

    @base_pth.setter
    def base_pth(self, base_pth):
        """Set the URI's extension property, after checking its conformity."""
        self.__base_pth = self.check_base_pth(base_pth)

    def normpath(func):
        """replace backslashes by forward slashes also on Windows
        This is necessary to make the doctests behave the same way
        on Windows, Mac and Unix systems"""
        def normalize(*args, **kwargs):
            r = func(*args, **kwargs)
            return re.sub(r"\\+","/", r.encode('unicode-escape').decode())
        return normalize

    @normpath
    def check_base_pth(self, base_pth):
        """Check whether the base_pth is None, in wich case it is set to '.'"""
        if base_pth == None:
            return "."
        return base_pth

    @property
    def uri_type(self):
        """Get the URI object's current uri_type
        ("author", "book", "version", None) based on its defined components.

        NB: uri_type does not have a setter method, making it read-only!

        Examples:
            >>> my_uri = URI("0255Jahiz.Hayawan.Sham19Y0023775-ara1.inProgress")
            >>> my_uri.uri_type
            'version'
            >>> my_uri.language = ""
            >>> my_uri.uri_type
            'book'
            >>> my_uri = URI()
            >>> my_uri.uri_type is None
            True
        """
        uri_type = None
        if self.date and self.author:
            uri_type = "author"
            if self.title:
                uri_type = "book"
                if self.version and self.language and self.edition_no:
                    uri_type = "version"
        return uri_type

    ############################################################################

    def __call__(self, uri_type=None, ext=None):
        """Call the self.build_uri() method of the URI instance.

        Examples:
            >>> my_uri = URI("0768IbnMuhammadTaqiDinBaclabakki")
            >>> my_uri.title = "Hadith"
            >>> my_uri.version = "Shamela0009426"
            >>> my_uri.language = "ara"
            >>> my_uri.edition_no = "1"
            >>> my_uri()
            '0768IbnMuhammadTaqiDinBaclabakki.Hadith.Shamela0009426-ara1'
            >>> my_uri("date")
            '0768'
            >>> my_uri("author")
            '0768IbnMuhammadTaqiDinBaclabakki'
            >>> my_uri("author_yml")
            '0768IbnMuhammadTaqiDinBaclabakki.yml'
            >>> my_uri("book")
            '0768IbnMuhammadTaqiDinBaclabakki.Hadith'
            >>> my_uri("book_yml")
            '0768IbnMuhammadTaqiDinBaclabakki.Hadith.yml'
            >>> my_uri("version")
            '0768IbnMuhammadTaqiDinBaclabakki.Hadith.Shamela0009426-ara1'
            >>> my_uri("version_yml")
            '0768IbnMuhammadTaqiDinBaclabakki.Hadith.Shamela0009426-ara1.yml'
            >>> my_uri("version_file", ext="completed")
            '0768IbnMuhammadTaqiDinBaclabakki.Hadith.Shamela0009426-ara1.completed'
        """
        return self.build_uri(uri_type, ext)


    def __repr__(self):
        """Return a representation of the components of the URI.

        Returns:
            (str): a representation of the components of the URI.

        Examples:
            >>> my_uri = URI("0255Jahiz.Hayawan.Sham19Y0023775-ara1")
            >>> repr(my_uri)
            'uri(date:0255, author:Jahiz, title:Hayawan, version:Sham19Y0023775, \
language:ara, edition_no:1, extension:)'
            >>> my_uri = URI()
            >>> repr(my_uri)
            'uri(date:, author:, title:, version:, language:, edition_no:, extension:)'
        """
        component_names = "date author title version language edition_no extension"
        fmt = [x+":{_URI__"+x+"}" for x in component_names.split()]
        fmt = "uri({})".format(", ".join(fmt))
        return fmt.format(**self.__dict__)

    def __str__(self, *args, **kwargs):
        """Return the reassembled URI.

        Returns:
            (str): the reassembled URI

        Examples:
            >>> my_uri = URI("0255Jahiz.Hayawan.Sham19Y0023775-ara1.inProgress")
            >>> my_uri.extension = "completed"
            >>> print(my_uri)
            0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed
        """
        return self.build_uri()


    def __iter__(self):
        """Enable iteration over a URI object.

        Returns:
            (iterator): an iterator containing the components of the URI

        Examples:
            >>> my_uri = URI("0255Jahiz.Hayawan.Sham19Y0023775-ara1.inProgress")
            >>> for component in my_uri: print(component)
            0255
            Jahiz
            Hayawan
            Sham19Y0023775
            ara
            1
            inProgress
            >>> my_uri = URI()
            >>> for component in my_uri: print(component)

        """
        return iter(self.split_uri())
##        return iter([self.date, self.author, self.title, self.version,
##                     self.language, self.edition_no, self.extension])


    ############################################################################


##    def get_uri_type(self):
##        """Get the type of the URI object.
##
##        Examples:
##            >>> my_uri = URI("0255Jahiz.Hayawan.Sham19Y0023775-ara1.inProgress")
##            >>> my_uri.get_uri_type()
##            'version'
##            >>> my_uri = URI()
##            >>> my_uri.get_uri_type() is None
##            True
##        """
##        uri_type = None
##        if self.date and self.author:
##            uri_type = "author"
##            if self.title:
##                uri_type = "book"
##                if self.version and self.language and self.edition_no:
##                    uri_type = "version"
##        return uri_type

    def split_uri(self, uri_string=None):
        """
        Split an OpenITI URI string into its components and check if components are valid.

        Args:
            uri_string (str): OpenITI URI, e.g.,
                0768IbnMuhammadTaqiDinBaclabakki.Hadith.Shamela0009426-ara1

        Returns:
            (list): list of uri components

        Examples:
            >>> my_uri = URI("0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed")
            >>> my_uri.split_uri()
            ['0255', 'Jahiz', 'Hayawan', 'Sham19Y0023775', 'ara', '1', 'completed']
            >>> my_uri.extension=""
            >>> my_uri.language=""
            >>> my_uri.split_uri()
            ['0255', 'Jahiz', 'Hayawan']
        """
        if not uri_string:
            uri_string = self.build_uri()
        split_uri = uri_string.split(".")

        if split_uri == [""]:
            return []

        if len(split_uri) > 4:
            msg = "URI ({}) has too many parts separated by dots".format(uri_string)
            raise Exception(msg)

        self.dateAuth = split_uri[0]
        self.date = re.findall("^\d+", self.dateAuth)[0]
        self.check_date(self.date)
        self.author = self.dateAuth[4:]
        if not self.author:
            msg = "No author name found. \
Did you put a dot between date and author name?"
            raise Exception(msg)
        split_components = [self.date, self.author]

        if len(split_uri) > 1:
            if split_uri[1] != "yml":
                self.title = split_uri[1]
                split_components.append(self.title)

        if len(split_uri) > 2:
            if split_uri[2] != "yml":
                self.versionLang = split_uri[2]
                try:
                    self.version, language = self.versionLang.split("-")
                except:
                    raise Exception("URI () misses language ")
                self.check_ASCII(self.version, "Version ID")
                split_components.append(self.version)
                if language[-1].isnumeric():
                    self.edition_no = re.findall("\d+", language)[0]
                    self.language = re.sub("\d+", "", language)
                    split_components.append(self.language)
                    split_components.append(self.edition_no)
                else:
                    self.edition_no = ""
                    self.language = language
                    split_components.append(self.language)
                self.check_language_code(self.language)
        if len(split_uri) > 3:
            #if split_uri[3] != "yml":
            self.extension = split_uri[3]
            split_components.append(self.extension)
            #else:
##                print("ERROR: extension '{}' not in the list\
##of acceptable extensions. No extension recorded".format(split_uri[3]))
            #    self.extension = ""
        return split_components


    def build_uri(self, uri_type=None, ext=None):
        """
        Build an OpenITI URI string from its components.

        Args:
            uri_type (str): the uri type to be returned (defaults to None):
                - "date" : only the date (format: 0000)
                - "author" : authorUri (format: 0255Jahiz)
                - "author_yml" : filename of the author yml file\
                  (format: 0255Jahiz.yml)
                - "book": BookUri (format: 0255Jahiz.Hayawan)
                - "book_yml": filename of the book yml file\
                  (format: 0255Jahiz.Hayawan.yml)
                - "version": versionURI\
                  (format: 0255Jahiz.Hayawan.Shamela000245-ara1)
                - "version_yml": filename of the version yml file\
                  (format: 0255Jahiz.Hayawan.Shamela000245-ara1.yml)
                - "version_file": filename of the version text file\
                  (format: 0255Jahiz.Hayawan.Shamela000245-ara1.completed)
            ext (str): extension for the version_file uri string
                (can be "completed", "inProgress", "mARkdown", "" or None).

        Returns:
            (str): OpenITI URI as a string, e.g.,
                0768IbnMuhammadTaqiDinBaclabakki.Hadith.Shamela0009426-ara1

        Examples:
            >>> my_uri = URI("0768IbnMuhammadTaqiDinBaclabakki")
            >>> my_uri.title = "Hadith"
            >>> my_uri.version = "Shamela0009426"
            >>> my_uri.language = "ara"
            >>> my_uri.edition_no = "1"
            >>> my_uri.build_uri()
            '0768IbnMuhammadTaqiDinBaclabakki.Hadith.Shamela0009426-ara1'
            >>> my_uri.build_uri("date")
            '0768'
            >>> my_uri.build_uri("author")
            '0768IbnMuhammadTaqiDinBaclabakki'
            >>> my_uri.build_uri("author_yml")
            '0768IbnMuhammadTaqiDinBaclabakki.yml'
            >>> my_uri.build_uri("book")
            '0768IbnMuhammadTaqiDinBaclabakki.Hadith'
            >>> my_uri.build_uri("book_yml")
            '0768IbnMuhammadTaqiDinBaclabakki.Hadith.yml'
            >>> my_uri.build_uri("version")
            '0768IbnMuhammadTaqiDinBaclabakki.Hadith.Shamela0009426-ara1'
            >>> my_uri.build_uri("version_yml")
            '0768IbnMuhammadTaqiDinBaclabakki.Hadith.Shamela0009426-ara1.yml'
            >>> my_uri.build_uri("version_file", ext="completed")
            '0768IbnMuhammadTaqiDinBaclabakki.Hadith.Shamela0009426-ara1.completed'
        """
        self.uri_string = ""

        if not uri_type:
            if self.version and self.language:
                if self.extension:
                    return self.build_uri("version_file", ext=ext)
                else:
                    return self.build_uri("version", ext=ext)
            elif self.title:
                return self.build_uri("book", ext=ext)
            elif self.author:
                return self.build_uri("author", ext=ext)
            elif self.date:
                return self.build_uri("date", ext=ext)
            else:
                return ""

        if uri_type == "date":
            if self.date:
                self.uri_string = str(self.date)
            else:
                raise Exception("Error: the date component of the URI was not defined")
        elif "author" in uri_type:
            if self.author:
                self.uri_string = "{}{}".format(self.build_uri("date"),self.author)
            else:
                raise Exception("Error: the author component of the URI was not defined")
        elif "book" in uri_type:
            if self.title:
                self.uri_string = "{}.{}".format(self.build_uri("author"), self.title)
            else:
                raise Exception("Error: the title component of the URI was not defined")
        elif "version" in uri_type:
            if self.version and self.language:
                self.uri_string =  "{}.{}-{}{}".format(self.build_uri("book"),
                                                       self.version,
                                                       self.language,
                                                       self.edition_no)
                if "file" in uri_type:
                    if ext != None:
                        if ext != "":
                            self.uri_string += ".{}".format(ext)
                        # else: do not add an extension
                    else:
                        if self.extension:
                            self.uri_string += ".{}".format(self.extension)
            elif self.version:
                raise Exception("Error: the language component of the URI was not defined")
            elif self.language:
                raise Exception("Error: the version component of the URI was not defined")
            else:
                raise Exception("Error: the language and version components of the URI were not defined")
        if "yml" in uri_type:
            self.uri_string += ".yml"
        return self.uri_string

    def get_version_uri(self):
        """
        Returns the version uri.

        Returns:
            (str): OpenITI URI as a string, e.g.,
                0768IbnMuhammadTaqiDinBaclabakki.Hadith.Shamela0009426-ara1

        Example:
            >>> my_uri = URI('0768IbnMuhammadTaqiDinBaclabakki.Hadith.Shamela0009426-ara1')
            >>> my_uri.extension = "completed"
            >>> my_uri.get_version_uri()
            '0768IbnMuhammadTaqiDinBaclabakki.Hadith.Shamela0009426-ara1'
        """
        return self.build_uri("version")

    def get_book_uri(self):
        """
        Returns the book uri.

        Returns:
            uri_string (str): OpenITI URI as a string, e.g.,
                0768IbnMuhammadTaqiDinBaclabakki.Hadith

        Example:
            >>> my_uri = URI('0768IbnMuhammadTaqiDinBaclabakki.Hadith.Shamela0009426-ara1')
            >>> my_uri.get_book_uri()
            '0768IbnMuhammadTaqiDinBaclabakki.Hadith'
        """
        return self.build_uri("book")

    def get_author_uri(self):
        """
        Returns the author uri.

        Returns:
            uri_string (str): OpenITI URI as a string, e.g.,
                0768IbnMuhammadTaqiDinBaclabakki

        Example:
            >>> my_uri = URI('0768IbnMuhammadTaqiDinBaclabakki.Hadith.Shamela0009426-ara1')
            >>> my_uri.get_author_uri()
            '0768IbnMuhammadTaqiDinBaclabakki'
        """
        return self.build_uri("author")


    @normpath  # always use "/" as path separator, for use across Unix and Windows
    def build_pth(self, uri_type=None, base_pth=None):
        """build the path to a file or folder using the OpenITI uri system

        Args:
            uri_type (str): the uri type of the path to be returned
            (defaults to None):
                - "date" : only the 25-years openITI folder (format: 0275AH)
                - "author" : authorUri (format: 0255Jahiz)
                - "author_yml" : path to the author yml file
                    (format: 0275AH/0255Jahiz.yml)
                - "book": path to the book folder
                    (format: 0275AH/0255Jahiz/0255Jahiz.Hayawan)
                - "book_yml": path to the book yml file
                    (format: 0275AH/0255Jahiz/0255Jahiz.Hayawan/0255Jahiz.Hayawan.yml)
                - "version": path to the book folder
                    in which the version files reside
                    (format: 0275AH/0255Jahiz/0255Jahiz.Hayawan)
                - "version_yml": path to the version yml file
                    (format: 0275AH/data/0255Jahiz/0255Jahiz.Hayawan/0255Jahiz.Hayawan.Sham19Y0023775-ara1.yml)
                - "version_file": path to the version text file
                    (format: 0275AH/data/0255Jahiz/0255Jahiz.Hayawan/0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed)
            base_pth (str): path to the root folder,
                to be prepended to the URI path

        Returns:
            (str): relative/absolute path

        Examples:
            >>> my_uri = URI("0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed")
            >>> my_uri.build_pth(base_pth="./master", uri_type="date")
            './master/0275AH'
            >>> my_uri.build_pth(base_pth="./master", uri_type="author")
            './master/0275AH/data/0255Jahiz'
            >>> my_uri.build_pth(base_pth="./master", uri_type="author_yml")
            './master/0275AH/data/0255Jahiz/0255Jahiz.yml'
            >>> my_uri.build_pth(base_pth="./master", uri_type="book")
            './master/0275AH/data/0255Jahiz/0255Jahiz.Hayawan'
            >>> my_uri.build_pth(base_pth="./master", uri_type="book_yml")
            './master/0275AH/data/0255Jahiz/0255Jahiz.Hayawan/0255Jahiz.Hayawan.yml'
            >>> my_uri.build_pth(base_pth="./master", uri_type="version_file")
            './master/0275AH/data/0255Jahiz/0255Jahiz.Hayawan/0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed'
             >>> my_uri.build_pth(base_pth="./master", uri_type="version")
            './master/0275AH/data/0255Jahiz/0255Jahiz.Hayawan'
       """
        if base_pth is None:
            base_pth = self.base_pth
        #print("base_pth:", base_pth)

        if not uri_type:
            if self.version and self.language:
                if self.extension:
                    return self.build_pth(uri_type="version_file", base_pth=base_pth)
                else:
                    return self.build_pth(uri_type="version", base_pth=base_pth)
            elif self.title:
                return self.build_pth(uri_type="book", base_pth=base_pth)
            elif self.author:
                return self.build_pth(uri_type="author", base_pth=base_pth)
            elif self.date:
                return self.build_pth("date", base_pth)


        if uri_type == "date":
            if self.date:
                if str(self.date).endswith("AH"):
                    self.date=self.date[:-2]
                if int(self.date)%25:
                    d = "{:04d}AH".format((int(int(self.date)/25) + 1)*25)
                else:
                    d = "{:04d}AH".format(int(self.date))
                #p = os.path.join(base_pth, d)
                return os.sep.join((base_pth, d))
            else:
                raise Exception("Error: the date component of the URI was not defined")
        elif "author" in uri_type:
            if self.data_in_25_year_repos: 
                pth = os.sep.join((self.build_pth("date", base_pth), "data",
                                   self.build_uri("author")))
            else:
                pth = os.sep.join((base_pth, self.build_uri("author")))
        elif "book" in uri_type:
            pth = os.sep.join((self.build_pth("author", base_pth),
                                self.build_uri("book")))
        elif "version" in uri_type:
            pth = self.build_pth("book", base_pth)
##            if "yml" in uri_type or "file" in uri_type:
##                pth = (self.build_pth("book", base_pth))
##            else:
##                pth = os.sep.join((self.build_pth("book", base_pth),
##                                   self.build_uri("version")))
        if "yml" in uri_type or "file" in uri_type:
            return pth + os.sep + self.build_uri(uri_type)
        else:
            return pth

    def from_folder(self, folder):
        """Create a URI from a folder path without a file name."""
        # not yet implemented



################################################################################
# OpenITI corpus functions dependent on URIs:


def change_uri(old, new, old_base_pth=None, new_base_pth=None, execute=False):
    """Change a uri and put all files in the correct folder.

    If a version URI changes:
    
        * all text files of that version should be moved
        * the yml file of that version should be updated and moved
        
    If a book uri changes:
    
        * the yml file of that book should be updated and moved
        * all annotation text files of all versions of the book should be moved
        * all yml files of versions of that book should be updated and moved
        * the original book folder itself should be (re)moved
        
    if an author uri changes:

        * the yml file of that author should be updated and moved
        * all book yml files of that should be updated and moved
        * all annotation text files of all versions of all books should be moved
        * all yml files of versions of all books should be updated and moved
        * the original book folders should be (re)moved
        * the original author folder itself should be (re)moved

    Examples::
    
        change_uri("0255Jahiz", "0256Jahiz")
        change_uri("0255Jahiz", "0255JahizBasri")
        change_uri("0255Jahiz.Hayawan", "0255Jahiz.KitabHayawan")
        change_uri("0255Jahiz.Hayawan.Shamela002526-ara1",\
                   "0255Jahiz.Hayawan.Shamela002526-ara2")
        change_uri("0255Jahiz.Hayawan.Shamela002526-ara1.completed",\
                   "0255Jahiz.Hayawan.Shamela002526-ara1.mARkdown")

    Args:
        old (str): URI string to be changed
        new (str): URI string to which the new URI should be changed.
        old_base_pth (str): path to the folder containing the
            OpenITI 25-year repos, related to the old uri
        new_base_pth (str): path to the folder containing the
            OpenITI 25-year repos, related to the new uri
        execute (bool): if False, the proposed changes will only be printed
            (the user will still be given the option to execute
            all proposed changes at the end);
            if True, all changes will be executed immediately.

    Returns:
        None
    """
    print("old_base_pth:", old_base_pth)
    old_uri = URI(old)
    old_uri.base_pth = old_base_pth
    new_uri = URI(new)
    new_uri.base_pth = new_base_pth

    if not execute:
        print("old uri:", old)
        print("new uri:", new)
        print("Proposed changes:")
    old_folder = old_uri.build_pth()
    if new_uri.uri_type == "version":
        # only move yml and text file(s) of this specific version:
        for file in os.listdir(old_folder):
            fp = os.path.join(old_folder, file)
            if not file.endswith(".md"):
                if URI(file).build_uri(ext="") == old_uri.build_uri(ext=""):
                    if file.endswith(".yml"):
                        move_yml(fp, new_uri, "version", execute)
                    else:
                        old_file_uri = URI(fp)
                        new_uri.extension = old_file_uri.extension
                        move_to_new_uri_pth(fp, new_uri, execute)

        # add readme and text_questionnaire files:

        target_folder = new_uri.build_pth("version")
        if execute:
            if "README.md" not in os.listdir(target_folder):
                add_readme(target_folder)
            if "text_questionnaire.md" not in os.listdir(target_folder):
                add_text_questionnaire(target_folder)
        else:
            print("add/move readme and text questionnaire files")

    else: # move all impacted files and directories
        for root, dirs, files in os.walk(old_folder):
            for file in files:
                if file in ["README.md", "text_questionnaire.md"]:
                    # skip README and text_questionnaire files until last
                    # so we can use the uri of the other files to create path
                    pass
                else:
                    print()
                    print("* file:", file)
                    fp = os.path.join(root, file)
                    old_file_uri = URI(fp)
                    print("  (type: {})".format(old_file_uri.uri_type))
                    new_file_uri = copy.deepcopy(old_file_uri)
                    new_file_uri.base_pth = new_uri.base_pth
                    new_file_uri.date = new_uri.date
                    new_file_uri.author = new_uri.author
                    if new_uri.uri_type == "book":
                        new_file_uri.title = new_uri.title
                    if file.endswith(".yml"):
                        new_fp = move_yml(fp, new_file_uri,
                                          old_file_uri.uri_type, execute)
                    else:
                        new_fp = move_to_new_uri_pth(fp, new_file_uri, execute)

            # Deal with non-URI filenames last:

            for fn in ["README.md", "text_questionnaire.md"]:
                fp = os.path.join(root, fn)
                new_folder = os.path.split(new_fp)[0] # defined in previous loop
                new_fp = os.path.join(new_folder, fn)
                if os.path.exists(fp):
                    # if the old folder contained a readme / text questionnaire:
                    if execute:
                        shutil.move(fp, new_fp)
                        print("  Moved", fp, "to", new_fp)
                    else:
                        print("  Move", fp, "to", new_fp)

    # Remove folders:

    if new_uri.uri_type == "author":
        if execute:
            shutil.rmtree(old_folder)
        else:
            for book_dir in os.listdir(old_folder):
                if not book_dir.endswith("yml"):
                    print("REMOVE BOOK FOLDER", os.path.join(old_folder, book_dir))
            print("REMOVE AUTHOR FOLDER", old_folder)
    if new_uri.uri_type == "book":
        if execute:
            shutil.rmtree(old_folder)
        else:
            print("REMOVE BOOK FOLDER", old_folder)

    if not execute:
        resp = input("To carry out these changes: press OK+Enter; \
to abort: press Enter. ")
        if resp == "OK":
            print()
            change_uri(old, new, old_base_pth, new_base_pth, execute=True)
        else:
            print("User aborted carrying out these changes!")

def add_readme(target_folder):
    """Add default README.md file to target_folder.

    Returns:
        None
    """
    with open(os.path.join(target_folder, "README.md"),
              mode="w", encoding="utf-8") as file:
        file.write(readme_template)

def add_text_questionnaire(target_folder):
    """Add default text_questionnaire.md file to target_folder.

    Returns:
        None
    """
    with open(os.path.join(target_folder, "text_questionnaire.md"),
              mode="w", encoding="utf-8") as file:
        file.write(text_questionnaire_template)

def add_character_count(tok_count, char_count, tar_uri, execute=False):
    """Add the character and token counts to the new version yml file

    Args:
        tok_count (int): number of Arabic tokens in a text
        char_count (int): number of Arabic characters in a text
        tar_uri (URI object): uri of the target text
        execute (bool): if True, the function will do its work silently.
          If False, it will only print a description of the action.

    Returns:
        None
    """

    tar_yfp = tar_uri.build_pth("version_yml")
    if execute:
        with open(tar_yfp, mode="r", encoding="utf-8") as file:
            yml_dic = yml.ymlToDic(file.read().strip())
            yml_dic["00#VERS#LENGTH###:"] = tok_count
            yml_dic["00#VERS#CLENGTH##:"] = char_count
        with open(tar_yfp, mode="w", encoding="utf-8") as file:
            file.write(yml.dicToYML(yml_dic))
    else:
        print("  Add the character count to the version yml file")


def new_yml(tar_yfp, yml_type, execute=False):
    """Create a new yml file from template.

    Args:
        tar_yfp (str): filepath to the new yml file
        yml_type (str): type of yml file
            (either "version_yml", "book_yml", or "author_yml")
    Returns:
        None
    """
    template = eval("{}_template".format(yml_type))
    yml_dic = yml.ymlToDic(template)
    uri_key = "00#{}#URI######:".format(yml_type[:4].upper())
    u = URI(tar_yfp)
    u.extension = ""
    yml_dic[uri_key] = u.build_uri()
    if execute:
        with open(tar_yfp, mode="w", encoding="utf-8") as file:
            file.write(yml.dicToYML(yml_dic))
    else:
        if not tar_yfp in created_ymls:
            print("  Create temporary yml file", tar_yfp)
            created_ymls.append(tar_yfp)


def move_yml(yml_fp, new_uri, uri_type, execute=False):
    """Replace the URI in the yml file
    and save the yml file in its new location.

    Args:
        yml_fp (str): path to the original yml file
        new_uri (URI object): the new uri
        uri_type (str): uri type (author, book, version)
        execute (bool): if False, the proposed changes will only be printed
            (the user will still be given the option to execute
            all proposed changes at the end);
            if True, all changes will be executed immediately.

    Returns:
        (str): filepath of the new yml file
    """
    new_yml_fp = new_uri.build_pth(uri_type=uri_type+"_yml")
    new_yml_folder = os.path.split(new_yml_fp)[0]
    make_folder(new_yml_folder, new_uri, execute)

    if not execute:
        print("  Change URI inside yml file:")
    yml_dict = yml.readYML(yml_fp)
    key = "00#{}#URI######:".format(uri_type[:4].upper())
    yml_dict[key] = new_uri.build_uri(uri_type=uri_type, ext="")
    yml_str = yml.dicToYML(yml_dict)
    if not execute:
        print(yml_str)

    if execute:
        with open(new_yml_fp, mode="w", encoding="utf-8") as yml_file:
            yml_file.write(yml_str)
        os.remove(yml_fp)
        print("Moved", yml_fp, "\n    to", new_yml_fp)
    else:
        print("  Move", yml_fp, "\n    to", new_yml_fp)
    return new_yml_fp


def make_folder(new_folder, new_uri, execute=False):
    """Check if folder exists; if not, make folder (and, if needed, parents)

    Args:
        new_folder (str): path to new folder
        new_uri (OpenITI uri object): uri of the text
        execute (bool): if False, the proposed changes will only be printed
            (the user will still be given the option to execute
            all proposed changes at the end);
            if True, all changes will be executed immediately.

    Returns:
        None
    """
    if not os.path.exists(new_uri.base_pth):
        msg = """PathError: base path ({}) does not exist.
Make sure base path is correct.""".format(new_uri.base_pth)
        raise Exception(msg)
    if not os.path.exists(new_folder):
        author_folder = new_uri.build_pth("author")
        if not os.path.exists(author_folder):
            if execute:
                os.makedirs(author_folder)
            else:
                if not author_folder in created_folders:
                    print("  Make author_folder", author_folder)
                    created_folders.append(author_folder)
            new_yml(new_uri.build_pth("author_yml"), "author_yml", execute)
        if new_uri.uri_type == "book" or new_uri.uri_type == "version":
            book_folder = new_uri.build_pth("book")
            if not os.path.exists(book_folder):
                if execute:
                    os.makedirs(book_folder)
                else:
                    if not book_folder in created_folders:
                        print(" Make book_folder", book_folder)
                        created_folders.append(book_folder)
                new_yml(new_uri.build_pth("book_yml"), "book_yml", execute)
            if new_uri.uri_type == "version":
                new_yml(new_uri.build_pth("version_yml"),
                            "version_yml", execute)
                target_folder = new_uri.build_pth("version")
                if execute:
                    if "README.md" not in os.listdir(target_folder):
                        add_readme(target_folder)
                    if "text_questionnaire.md" not in os.listdir(target_folder):
                        add_text_questionnaire(target_folder)


def move_to_new_uri_pth(old_fp, new_uri, execute=False):
    """Move file to its new location.

    Args:
        old_fp (filepath): path to the old file
        new_uri (URI object): URI of the new file
        execute (bool): if False, the proposed changes will only be printed
            (the user will still be given the option to execute
            all proposed changes at the end);
            if True, all changes will be executed immediately.

    Returns:
        (str): path to the new file
    """
    new_folder = new_uri.build_pth(uri_type=new_uri.uri_type)
    new_fp = new_uri.build_pth(uri_type=new_uri.uri_type+"_file")
    make_folder(new_folder, new_uri, execute)
    if execute:
        shutil.move(old_fp, new_fp)
        print("  Move", old_fp, "\n    to", new_fp)
    else:
        print("  Move", old_fp, "\n    to", new_fp)
    return new_fp


#def check_token_count(version_uri, ymlD):
def check_token_count(version_uri, ymlD, version_fp="", find_latest=True):
    """Check whether the token count in the version yml file agrees with the\
    actual token count of the text file.

    Args:
        version_uri (URI object): version uri of the target text
        ymlD (dict): dictionary containing the data from the relevant yml file
        version_fp (str): file path to the target text
        find_latest (bool): if False, the version_fp will be used as is;
            if set to True, the script will find the most developed version of
            the text file, based on its extension (mARkdown > completed > inProgress)
    Returns:
        (tuple): Tuple containing 2 values (or None):

            tok_count (int): number of Arabic tokens in the target text
            char_count (int): number of Arabic characters in the target text
    """
    # Get the count from the most complete version of the text file: 
    #fp = version_uri.build_pth(uri_type="version_file")
    if version_fp and not find_latest:
        fp = version_fp
    elif version_fp and find_latest:
        for ext in [".mARkdown", ".completed", ".inProgress", ""]:
            fp = version_fp + ext
            if os.path.exists(fp):
                break
    else:
        for ext in ["mARkdown", "completed", "inProgress", ""]:
            version_uri.extension = ext
            fp = version_uri.build_pth(uri_type="version_file")
            if os.path.exists(fp):
                break
    tok_count = ar_cnt_file(fp, mode="token")
    char_count = ar_cnt_file(fp, mode="char")
    len_key = "00#VERS#LENGTH###:"
    char_len_key = "00#VERS#CLENGTH##:"
    yml_tok_count = ymlD[len_key].strip()
    try:
        yml_char_count = ymlD[char_len_key].strip()
    except:
        yml_char_count = ""
    replace_tok_count = False
    for cnt, yml_cnt in [(tok_count, yml_tok_count),
                         (char_count, yml_char_count)]:
        if yml_cnt == "":
            print("NO TOKEN COUNT", version_uri)
            replace_tok_count = True
        else:
            try:
                if int(yml_cnt) != cnt:
                    replace_tok_count = True
                    #print("TOKEN COUNT CHANGED", uri)
                    #print(yml_tok_count, "!=", tok_count)
            except:
                print("TOKEN COUNT {} IS NOT A NUMBER".format(yml_cnt), uri)
                replace_tok_count = True
    if replace_tok_count:
        return tok_count, char_count

def replace_tok_counts(missing_tok_count):
    """Replace the token counts in the relevant yml files.

    Args:
        missing_tok_count (list): a list of tuples:
            uri (OpenITI URI object)
            version_fp (str)
            token_count (int): the number of Arabic tokens in the text file
            char_count (int): the number of Arabic characters in the text file

    Returns:
        None
    """
    print("replacing token count in {} files".format(len(missing_tok_count)))
    for uri, version_fp, tok_count, char_count in missing_tok_count:
        print(uri, version_fp)
        #yml_fp = uri.build_pth("version_yml")
        if version_fp.endswith(("mARkdown", "completed", "inProgress")):
            yml_fp = os.path.splitext(version_fp)[0] + ".yml"
        else:
            yml_fp = version_fp + ".yml"
        ymlD = yml.readYML(yml_fp)
        len_key = "00#VERS#LENGTH###:"
        ymlD[len_key] = str(tok_count)
        char_len_key = "00#VERS#CLENGTH##:"
        ymlD[char_len_key] = str(char_count)
        ymlS = yml.dicToYML(ymlD)
        with open(yml_fp, mode="w", encoding="utf-8") as outf:
            outf.write(ymlS)

def check_yml_file(yml_fp, yml_type, version_fp=None, execute=False,
                   check_token_counts=True):
    """Check whether a yml file exist, is valid, and contains no foreign keys

    Args:
        yml_fp (str): path to the yml file
        yml_type (str): either "author", "book", or "version"
        version_fp (str): path to the text file of the version
            (only relevant for version yml files; default = None)
        execute (bool): if False, the user will be prompted
            before any changes are made to the yml file
        check_token_counts (bool): if True, the script will check
            the number of tokens (and characters) in the text

    Returns:
        None or yml_fp
    """
    yml_changed = False
    
    # Check if yml file exists:
    if not os.path.exists(yml_fp):
        print(yml_fp, "DOES NOT EXIST")
        # create a new yml file:
        if execute or input("Create yml file? Y/N? ").lower() == "y":
            new_yml(yml_fp, yml_type+"_yml", True)
            print("New yml file created.")
            yml_d = yml.readYML(yml_fp) 
        else:
            print("No new yml file created. Check manually!")
            return yml_fp
        

    # Check if yml is valid:
    try:
        yml_d = yml.readYML(yml_fp)
        if yml_d == {}:
            print("Empty yml file")
            if execute or input("Create yml file? Y/N? ").lower() == "y":
                new_yml(yml_fp, yml_type+"_yml", True)
                print("New yml file created.")
                yml_d = yml.readYML(yml_fp)
            else:
                print("No new file created. Check manually!")
                return yml_fp
            
        yml_d.keys()
    except Exception as e:
        print("invalid YML file structure:", yml_fp)
        print("Error message:", e)
        yml_d = yml.fix_broken_yml(yml_fp, execute)
        if yml_d:
            yml_changed = True
        else:
            return yml_fp
    for key in list(yml_d.keys()):  # NB: list needed because otherwise keys cannot be deleted!
        
        # check if all keys have the prefix of the yml type (..#AUTH, ..#BOOK, ..#VERS):
        if key[3:7] != yml_type.upper()[:4]:
            print("wrong key in yml file", yml_fp, ":", key)
            if execute or input("Delete yml key {}? Y/N: ".format(key)).lower() == "y":
                del yml_d[key]
                yml_changed = True
                print("-> deleted yml key", key)
            else:
                return yml_fp

        # check whether the URI in the yml file is identical with that in the filename:
        if "URI" in key:
            fn = os.path.splitext(os.path.split(yml_fp)[-1])[0]
            fn = re.sub("\.inProgress|\.mARkdown|\.completed", "", fn)
            if yml_d[key].strip() != fn:
                print("URI", yml_d[key], "!= filename", fn)
                if execute or input("Replace URI with filename? Y/N: ").lower() == "y":
                    yml_d[key] = fn
                    yml_changed = True
                    print("-> URI replaced with", fn)
                else:
                    return yml_fp
                
    # check whether version yml files contain token and character length values:
    if yml_type == "version" and check_token_counts:
        res = check_token_count(URI(yml_fp), yml_d, version_fp)
        if res:
            tok_count, char_count = res
            if execute or input("Change token count? Y/N? ").lower() == "y":
                yml_d["00#VERS#LENGTH###:"] = str(tok_count)
                yml_d["00#VERS#CLENGTH##:"] = str(char_count)
                yml_changed = True
                print(yml_fp)
                print("-> token and character counts changed")
            else:
                return yml_fp

    # save changes to yml file if anything has changed: 
    if yml_changed:
        with open(yml_fp, mode="w", encoding="utf-8") as file:
            file.write(yml.dicToYML(yml_d))


def check_yml_files(start_folder, exclude=[],
                    execute=False, check_token_counts=True,
                    flat_folder=False):
    """Check whether yml files are missing or have faulty data in them.

    Args:
        start_folder (str): path to the parent folder of the folders
            that need to be checked.
        exclude (list): a list of directory names that should be excluded.
        execute (bool): if execute is set to False, the script will only show
            which changes it would undertake if set to True.
            After it has looped through all files and folders, it will give
            the user the option to execute the proposed changes.

    Returns:
        None
    """
    failed = []
    for fp in get_all_text_files_in_folder(start_folder, excluded_folders=exclude):
        uri = URI(fp)
        for yml_type in ("author", "book", "version"):
            yml_fn = uri.build_uri(uri_type="{}_yml".format(yml_type))
            if yml_type == "author":
                if flat_folder:
                    yml_fp = os.path.join(os.path.dirname(fp), yml_fn)
                else:
                    yml_fp = os.path.join(os.path.dirname(os.path.dirname(fp)), yml_fn)
            else:
                yml_fp = os.path.join(os.path.dirname(fp), yml_fn)
            r = check_yml_file(yml_fp, yml_type, version_fp=fp, execute=execute,
                               check_token_counts=check_token_counts)
            if r:
                failed.append(r)
    if failed:
        print("The following yml files could not be read. Please correct them manually:")
        for yml_fp in failed:
            print("*", yml_fp)
        print()


if __name__ == "__main__":
    import doctest
    doctest.testmod()
    print("passed doctests")
    print()

    folder = r"D:\London\OpenITI\25Y_repos/0450AH"
    check_yml_files(folder)
    input("CONTINUE?")

    # Additional tests:


##    URI.base_pth = r"."
##    print("URI path in 25Y-repo format:")
##    my_uri = URI("0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed")
##    print(my_uri.build_pth())
##    print("URI path in release format (without 25Y repos):")
##    URI.data_in_25_year_repos = False
##    my_uri = URI("0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed")
##    print(my_uri.build_pth())
##    print("URI path in 25Y-repo format:")
##    URI.data_in_25_year_repos = True
##    print(my_uri.build_pth())
##    my_uri.data_in_25_year_repos = False
##    print("URI path in release format (without 25Y repos):")
##    print(my_uri.build_pth())
##
##    # Reset:
##    URI.data_in_25_year_repos = True
##    URI.base_pth = r"D:\London\OpenITI\25Y_repos"
##
##    input("Press Enter to continue")
##    print()
    
##    exclude = (["OpenITI.github.io", "Annotation", "maintenance", "i.mech00",
##                "i.mech01", "i.mech02", "i.mech03", "i.mech04", "i.mech05",
##                "i.mech06", "i.mech07", "i.mech08", "i.mech09", "i.logic",
##                "i.cex", "i.cex_Temp", "i.mech", "i.mech_Temp", ".git"])
##    resp = check_yml_files(r"D:\London\OpenITI\25Y_repos",
##                           exclude=exclude, execute=False)
##    missing_ymls, missing_tok_count, non_uri_files, erratic_ymls = resp
##    #print(non_uri_files)
##    input("continue?")
##
    base_pth = r"D:\London\OpenITI\python_library\openiti\openiti\test"

##    # test initialize_new_texts_in_folder function:
##    barzakh = os.path.join(base_pth, "barzakh")
##    initialize_new_texts_in_folder(barzakh, base_pth, execute=True)
##    print("Texts in test/barzakh initialized; check if initialization was successful!")
##    input("Press Enter to continue")

##    # test initialize_texts_from_CSV function:
##    csv_fp = os.path.join(base_pth, "initialize.csv")
##    initialize_texts_from_CSV(csv_fp, old_base_pth="", new_base_pth=base_pth,
##                              execute=False)
##    print("Texts in test/barzakh initialized; check if initialization was successful!")
##    input("Press Enter to continue")

##    # test change_uri function for author uri change:
##    old = "0001KitabAllah"
##    new = "0001Allah"
##    change_uri(old, new,
##               old_base_pth=base_pth, new_base_pth=base_pth,
##               execute=False)
##    input("Press Enter to continue")

##    # test change_uri function for book uri change:
##    old = "0375IkhwanSafa.RisalatJamicaJamica"
##    new = "0375IkhwanSafa.RisalatJamica"
##    change_uri(old, new,
##               old_base_pth=base_pth, new_base_pth=base_pth,
##               execute=False)
##    input("Press Enter to continue")

##    # test change_uri function for version uri change:
##    old = "0001Allah.KitabMuqaddas.BibleCorpus002-per1"
##    new = "0001Allah.KitabMuqaddas.BibleCorpus2-per1"
##    change_uri(old, new,
##               old_base_pth=base_pth, new_base_pth=base_pth,
##               execute=False)
##    print("Check if the URI really changed??")
##    input("Press Enter to continue")

    my_uri = "0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed"

    #URI.base_pth = "XXXX"

    t = URI(my_uri)
    print("repr(t):")
    print(repr(t))
    print("print(t):")
    print(t)
    print("t.author:", t.author)
    print("t.date:", t.date)
    print("URI type:", t.uri_type)
    print('t.build_uri("author"):')
    print(t.build_uri("author"))
    print('t.build_pth("version_file"):')
    print(t.build_pth("version_file"))
    print('t.build_pth("version_file", ""):')
    print(t.build_pth("version_file", ""))

    print("*"*30)

    u = URI()
    u.author="IbnCarabi"
    u.date="0681"
    print('u.build_uri("author"):')
    print(u.build_uri("author"))
    print('u:')
    print(u)

    print("*"*30)

    my_uri = URI("0255Jahiz.Hayawan.Sham19Y0023775-ara1.completed")
    print('my_uri.split_uri():')
    print(my_uri.split_uri())
#    my_uri.extension=""
    my_uri.language=""
    print('my_uri.split_uri(): (language="")')
    print(my_uri.split_uri())
    my_uri.extension=""
    print("try to build a path from an incomplete version URI (language=''):")
    print("AN ERROR WARNING SHOULD FOLLOW: ")
    print(my_uri.build_pth(base_pth="./master", uri_type="version_file"))







"""Generic converter that converts HTML files to OpenITI mARkdown.

This generic converter forms the basis for specific converters
tailored to libraries whose texts are in html format.

Examples:
    Generic html conversion:
    >>> from html_converter_generic import GenericHtmlConverter
    >>> gen_converter = GenericHtmlConverter(dest_folder="test/converted")
    >>> gen_converter.VERBOSE = False
    >>> folder = r"test/eShia"
    >>> fn = "86596.html"
    >>> gen_converter.convert_file(os.path.join(folder, fn))
    >>> gen_converter.convert_files_in_folder(folder, extensions=[".html"])

    Sub-class to create a converter specific to htmls from the eShia library:
    >>> eShiaConv = GenericHtmlConverter("test/converted")
    >>> eShiaConv.VERBOSE = False
    >>> from helper import html2md_eShia
    >>> eShiaConv.add_structural_annotations = html2md_eShia.markdownify  # (1)
    >>> folder = r"test/eShia"
    >>> fn = "86596.html"
    >>> eShiaConv.convert_file(os.path.join(folder, fn))
    
    # (1) overwrite the add_structural_annotations method with another function

The GenericHtmlConverter is a subclass of the GenericConverter
from the generic_converter module:

GenericConverter
    \_ GenericEpubConverter
    \_ GenericHtmlConverter

GenericHtmlConverter's main methods are inherited from the GenericConverter:

* convert_file(source_fp): basic procedure for converting an html file.
* convert_files_in_folder(source_folder): convert all html files in the folder
    (calls convert_file)

Methods of both classes:
(methods of GenericConverter are inherited by GenericHtmlConverter;
methods of GenericConverter with the same name
in GenericHtmlConverter are overwritten by the latter)

=================================== ==========================
GenericConverter                    GenericHtmlConverter
=================================== ==========================
__init__                            __init__
convert_files_in_folder             (inherited)
convert file                        (inherited)
make_dest_fp                        (inherited - generic!)
get_metadata (dummy)                (inherited - dummy!)
get_data                            get_data
pre_process                         (inherited)
add_page_numbers (dummy)            (inherited - dummy!)
add_structural_annotations (dummy)  add_structural_annotations
remove_notes (dummy)                remove_notes
reflow                              (inherited)
add_milestones (dummy)              (inherited - dummy!)
post_process                        (inherited - generic!)
compose                             (inherited)
save_file                           (inherited)
                                    inspect_tags_in_html
                                    inspect_tags_in_folder
                                    find_example_of_tag
=================================== ==========================

The main difference between the two converters
is the add_structural_annotations method.
GenericHtmlConverter uses here the html2md converter
that converts or strips html tags.

To create a converter for a specific type of html files,
subclass the GenericHtmlConverter and overwrite
some of its methods that are dependent on the structure of the data, esp.:

- get_metadata
- add_page_numbers
- remove_notes
- add_structural_annotation (subclass the generic html2md converter for this,
  and create a html2md converter specific to the files you need to convert)

GenericConverter
    \_ GenericHtmlConverter
            \_ eShiaHtmlConverter
            \_ NoorlibHtmlConverter
            \_ ...

In addition to the conversion methods, the GenericHtmlConverter
contains a number of useful methods that help with deciding which
html tags need to be converted:

- inspect_tags_in_html
- inspect_tags_in_folder
- find_example_of_tag

"""

import codecs
import os
import re
import zipfile
import time

from bs4 import BeautifulSoup


if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(root_folder))
    sys.path.append(root_folder)

from openiti.new_books.convert.generic_converter import GenericConverter
from openiti.new_books.convert.helper import html2md



class GenericHtmlConverter(GenericConverter):

    def __init__(self, dest_folder=None):
        """Initialize the class and its super-class."""
        super().__init__()
        self.root_tag_string = "soup"
        self.unique_tags = dict()
##        if dest_folder == None:
##            self.dest_folder = "converted"
##        else:
##            self.dest_folder = dest_folder
        self.VERBOSE = True
        self.extension = ""


            
##    def convert_file(self, source_fp):
##        """Convert one file to OpenITI format.
##
##        Args:
##            source_fp (str): path to the file that must be converted.
##
##        Returns:
##            None
##        """
##        if self.VERBOSE:
##            print("converting", source_fp)
##        dest_fp = self.make_dest_fp(source_fp)
##        metadata = self.get_metadata(source_fp)
##
##        text = self.get_data(source_fp)
##        #print(1, text)
##        text = self.pre_process(text)
##        #print(2, text)
##
##        text = self.add_page_numbers(text, source_fp)
##        
##        text, notes = self.remove_notes(text)
##        #print(5, text)
##        
##        text = self.convert_html2md(text)
##        #text = self.add_structural_annotations(text)
##        #print(4, text)
##        text = self.reflow(text)
##        #print(6, text)
##        #text = self.add_milestones(text)
##        #print(7, text)
##
##        text = self.post_process(text)
##        #print(8, text)
##
##        text = self.compose(metadata, text, notes)
##        #print(9, text)
##        
##
##        self.save_file(text, dest_fp)
##        #print("saved text to", dest_fp)


    #def convert_html2md(self, html):
    def add_structural_annotations(self, text):
        """Convert the html to mARkdown text.

        *** USING A GENERIC HTML TO MARKDOWN CONVERTER HERE.
        OVERWRITE THIS FUNCTION IN THE SUB-CLASS
        TO CONVERT THE LIBRARY'S SPECIFIC HTML FLAVOUR TO MARKDOWN ***
        """
        text = html2md.markdownify(text)
        return text

    def remove_notes(self, text):
        """Remove the footnotes from the text.

        Args:
            text (str)

        Returns:
            text_without_notes (str): text from which the notes were removed.
            notes (str): the footnotes converted to endnotes
                (if there are endnotes: add the endnote_splitter before them)
        """
        text_without_notes = ""
        notes = ""
        spl = re.split("(FOOTNOTE.*\n+)", text)
        spl = [x for x in spl if x != ""]
        i = 0
        #last_text_fragm = 0
        for t in spl:
            if not t.strip().startswith("FOOTNOTE"):
                text_without_notes += t
                #last_text_fragm = i
            else:
                fn_text = t.strip()[8:]
                if spl[i-1].strip().startswith("FOOTNOTE"):
                    notes += "{}\n\n".format(fn_text)
                else:
                    #p = re.findall("PageV\d+P\d+", spl[last_text_fragm])
                    p = re.findall("PageV\d+P\d+", spl[i-1])
                    if p:
                        notes += "{}:\n\n{}\n\n".format(p[-1], fn_text)
                    else:
                        section_rgx = "(### [\|$]+) (.*)"
                        #sections = re.findall(section_rgx, spl[last_text_fragm])
                        sections = re.findall(section_rgx, spl[i-1])
                        if sections:
                            msg = "Notes on section ({})"
                            msg +=  "(no page numbers):\n\n{}\n\n"
                            notes += msg.format(sections[0][1], fn_text)
                        else:
                            notes += "(no page number):\n\n{}\n\n".format(fn_text)
            i += 1
        if notes:
            notes = self.endnote_splitter + notes
        return text_without_notes, notes



    def inspect_tags_in_html(self, html_str,
                             header_tag="",
                             only_tag_names=True,
                             strip_attr=["id", "href"],
                             print_tags=True):
        """Extracts a set of all tags used in an html string html_str.

        Args:
            html_str (str): a string in html format. Does not need to be well-formed.
            header_tag (str): the name of the tag that includes the header (e.g., "header");
                if defined, the tags within the header tag will not be taken into account.
                Defaults to an empty string (=> all tags in the file will be extracted)
            only_tag_names (bool): if set to True, only the tag names will be extracted,
                not the tag attributes. If set to False, the full tags including their
                attributes will be extracted. Defaults to True.
            strip_attr (list): the tag attributes in the strip_attr list will be
                stripped of their values.
                By default, the values of the "id" and "href" attributes will be stripped.
            print_tags(bool): if set to True, all single tags used in the folder will be printed.

        Returns:
            tags (set): a set of all tags that comply with the user's parameters.
        """
        if header_tag:
            header_tag = re.sub("\W", "", header_tag)
            html_str = re.sub(".+</{}>".format(header_tag), "", html_str,
                          count=1, flags=re.DOTALL)        
        if only_tag_names:
            tags = set(re.findall("<[^/][^ >]+", html_str))
        else:
            tags = set(re.findall("<[^/][^>]*>", html_str))

        if strip_attr and not only_tag_names:
            regex = '({})="[^"]+"'.format("|".join(strip_attr))
            stripped_tags = [re.sub(regex, r'\1=""', tag) for tag in tags]
            tags = set(stripped_tags)

        if print_tags:
            print("list of tags:")
            for tag in sorted(tags):
                print("    ", tag)
        return tags


    def inspect_tags_in_folder(self, folder,
                               extensions=[],
                               exclude_extensions=[],
                               header_tag="",
                               only_tag_names=True,
                               strip_attr=["id", "href"],
                               print_tags=True,
                               within_tag=None):
        """Extracts a set of all tags used in the texts in a folder.

        Args:
            folder (str): path to the folder from which the tags must be extracted.
            extensions (list): if defined, tags will be extracted
                only from files with one of these extensions.
                Defaults to an empty list (=> tags will be extracted from all files)
            exclude_extensions (list): if defined (and no extensions list is defined),
                tags will be extracted only from files without one of these extensions
                Defaults to an empty list (=> tags will be extracted from all files)
            header_tag (str): the name of the tag that includes the header (e.g., "header");
                if defined, the tags within the header tag will not be taken into account.
                Defaults to an empty string (=> all tags in the file will be extracted)
            only_tag_names (bool): if set to True, only the tag names will be extracted,
                not the tag attributes. If set to False, the full tags including their
                attributes will be extracted. Defaults to True.
            strip_attr (list): the tag attributes in the strip_attr list will be
                stripped of their values.
                By default, the values of the "id" and "href" attributes will be stripped.
            print_tags(bool): if set to True, all single tags used in the folder will be printed.

        Returns:
            tags (set): a set of all tags that comply with the user's parameters.
        """
        tags = set()
        fp_list = self.filter_files_in_folder(folder, extensions,
                                               exclude_extensions)
        for fp in fp_list: 
            print(fp)
            with open(fp, mode="r", encoding="utf-8") as file:
                text = file.read()
                if within_tag:
                    soup = BeautifulSoup(text)
                    text = soup.find(within_tag[0], attrs=within_tag[1]).prettify()
            fp_tags = self.inspect_tags_in_html(text, header_tag, only_tag_names, strip_attr,
                                                print_tags=False)
            tags = tags.union(fp_tags)

        if print_tags:
            print("list of tags:")
            for tag in sorted(tags):
                print("    ", tag)
        return tags

    def find_example_of_tag(self, folder, tag_name, attr={}, n=5, incl_parent=False,
                            extensions=[], exclude_extensions=[]):
        """Find an example of a specific tag from a file in the folder \
        with the required extension

        Args:
            folder (str): the path to the folder in which the files reside
            tag_name (str): the name of the tag for which an example is desired
            attr (dictionary): attribute of the tag for which an example is desired
                (key: attribute_name, value: attribute_value).
                E.g., {"class": "footnote"}
            n (int): number of examples wished
            incl_parent (bool): if True, the parent and siblings of the desired tag
                will be included.
            extensions (list): if defined, tags will be extracted
                only from files with one of these extensions.
                Defaults to an empty list (=> tags will be extracted from all files)
            exclude_extensions (list): if defined (and no extensions list is defined),
                tags will be extracted only from files without one of these extensions
                Defaults to an empty list (=> tags will be extracted from all files)
        """
        
        fp_list = self.filter_files_in_folder(folder, extensions=extensions,
                                              exclude_extensions=exclude_extensions)
        tags = []
        for fp in fp_list:
            with open(fp, mode="r", encoding="utf-8") as file:
                text = file.read()
            soup = BeautifulSoup(text)
            if attr:
                text_tags = soup.find_all(tag_name, attrs=attr)
            else:
                text_tags = soup.find_all(tag_name)
            if text_tags:
                print(fp)
            tags += text_tags
            if len(tags) >= n:
                break

        for t in tags[:n]:
            if incl_parent:
                t = t.parent
            print(t.prettify())
            print()
        return text_tags[:n]
        

if __name__ == "__main__":
    import doctest
    #doctest.testmod()
    print("Passed all tests.")

    conv = GenericHtmlConverter()
    folder = r"G:\London\OpenITI\new\Rafed\test"
##    conv.inspect_tags_in_folder(folder,
##                                extensions=["html"],
##                                only_tag_names=False,
##                                strip_attr=["name", "href", "title"],
##                                print_tags=True,
##                                within_tag=("div", {"id": "content1"}))

    conv.find_example_of_tag(folder, "p", attr={"class", "rfdCenterBold2"},
                             n=15, incl_parent=True, extensions=["html"])
    
##    conv.convert_files_in_folder(folder)






"""Converter that converts HTML files from the eShia library to OpenITI mARkdown.

The converter has two main functions:
* convert_file: convert a single html file.
* convert_files_in_folder: convert all html files in a given folder

Usage examples:
    >>> from html_converter_eShia import convert_file
    >>> folder = r"test/eShia/"
    >>> convert_file(folder+"86596.html", dest_fp=folder+"converted/86596")
    >>> from html_converter_eShia import convert_files_in_folder
    >>> convert_files_in_folder(folder, dest_folder=folder+"converted")  

Both functions use the EShiaHtmlConverter class to do the heavy lifting.
The EShiaHtmlConverter is a subclass of GenericHtmlConverter,
which in its turn inherits many functions from the GenericConverter.

GenericConverter
    \_ GenericHtmlConverter
            \_ EShiaHtmlConverter
            \_ NoorlibHtmlConverter
            \_ ...

Overview of the methods of these classes:
(methods of GenericConverter are inherited by GenericHtmlConverter;
and methods of GenericHtmlConverter are inherited by EShiaHtmlConverter.
Methods of the superclass with the same name
in the subclass are overwritten by the latter)

================================== ========================== ==========================
GenericConverter                   GenericHtmlConverter       EShiaHtmlConverter
================================== ========================== ==========================
__init__                           __init__                   (inherited)
convert_files_in_folder            (inherited)                (inherited)
convert file                       (inherited)                (inherited)
make_dest_fp                       (inherited - generic!)     (inherited - generic!)
get_metadata (dummy)               (inherited - dummy!)       get_metadata
get_data                           get_data                   (inherited)
pre_process                        (inherited)                (inherited)
add_page_numbers (dummy)           (inherited - dummy!)       add_page_numbers
add_structural_annotations (dummy) add_structural_annotations add_structural_annotations
remove_notes (dummy)               remove_notes               remove_notes
reflow                             (inherited)                (inherited)
add_milestones (dummy)             (inherited - dummy!)       (inherited - dummy!)
post_process                       (inherited - generic!)     post_process
compose                            (inherited)                (inherited)
save_file                          (inherited)                (inherited)
                                   inspect_tags_in_html       (inherited)
                                   inspect_tags_in_folder     (inherited)
                                   find_example_of_tag        (inherited)
================================== ========================== ==========================

The EShiaHtmlConverter's add_structural_annotations method uses html2md_eShia,
an adaptation of the generic html2md (based on markdownify)
to convert the html tags to OpenITI annotations. 

Examples:
    >>> from html_converter_eShia import EShiaHtmlConverter
    >>> conv = EShiaHtmlConverter()
    >>> conv.VERBOSE = False
    >>> folder = r"test/eShia/"
    >>> conv.convert_file(folder+"86596.html")
    >>> conv.convert_files_in_folder(folder, extensions=["html"])
"""

from bs4 import BeautifulSoup
import re

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(root_folder))
    sys.path.append(root_folder)

from openiti.new_books.convert.html_converter_generic import GenericHtmlConverter
from openiti.new_books.convert.helper import html2md_eShia


def convert_file(fp, dest_fp=None, verbose=False):
    """Convert one file to OpenITI format.

    Args:
        source_fp (str): path to the file that must be converted.
        dest_fp (str): path to the converted file.

    Returns:
        None
    """
    conv = EShiaHtmlConverter()
    conv.VERBOSE = verbose
    conv.convert_file(fp, dest_fp=dest_fp)

def convert_files_in_folder(src_folder, dest_folder=None,
                            extensions=["html"], exclude_extensions=["yml"],
                            fn_regex=None, verbose=False):
    """Convert all files in a folder to OpenITI format.\
    Use the `extensions` and `exclude_extensions` lists to filter\
    the files to be converted.

    Args:
        src_folder (str): path to the folder that contains
            the files that must be converted.
        dest_folder (str): path to the folder where converted files
            will be stored.
        extensions (list): list of extensions; if this list is not empty,
            only files with an extension in the list should be converted.
        exclude_extensions (list): list of extensions;
            if this list is not empty,
            only files whose extension is not in the list will be converted.
        fn_regex (str): regular expression defining the filename pattern
            e.g., "-(ara|per)\d". If `fn_regex` is defined,
            only files whose filename matches the pattern will be converted.


    Returns:
        None
    """
    conv = EShiaHtmlConverter()
    conv.VERBOSE = verbose
    conv.convert_files_in_folder(src_folder, dest_folder=dest_folder,
                                 extensions=extensions,
                                 exclude_extensions=exclude_extensions,
                                 fn_regex=fn_regex)


################################################################################


class EShiaHtmlConverter(GenericHtmlConverter):

##    # moved this function to the generic_converter!
##    def pre_process(self, text):
##        text = super().pre_process(text)
##        # attach separated wa- and a- prefixes to the following word: 
##        text = re.sub(r"\b([وأ])[\s~]+", r"\1", text)
##        return text

    def get_metadata(self, text):
        """Gets the metadata from the first td in the html.

        The metadata in eShia texts are in the first td
        (which represents the title page of the book).
        In some cases, the metadata is nicely parsed
        (p tags with ids like "author", "translator", ...);
        more often, they are simply in h, p or other tags,
        or only divided with <br/> tags.

        This function removes the footnotes from the title page
        and then adds the #META# tag before each line in the title page.        

        Args:
            text (str): text of the file that contains both data and metadata

        Returns:
            metadata (str): metadata formatted in OpenITI format
                (including the magic value and the header splitter)
        """
        soup = BeautifulSoup(text)
        meta_td = soup.find("td") # metadata is in first td tag
        
        # remove footnotes from metadata page:
        [x.extract() for x in meta_td.find_all("span", "footnotes")]
        [x.extract() for x in meta_td.find_all("footnote")]

        # remove all tags in the metadata page and add #META# tags: 
        metadata = re.sub("<[^>]+> *", "\n", str(meta_td))
        metadata = re.sub("\n+", "\n#META# ", metadata)

        # remove superfluous #META# tags:
        metadata = re.sub("\n* *#META# *(?:\n+|\Z)", "\n", metadata)

        # add magic value and header splitter: 
        metadata = self.magic_value + metadata + self.header_splitter
        return metadata

    def add_page_numbers(self, text, source_fp):
        """Convert the page numbers in the text into OpenITI mARkdown format"""
        try:
            vol_no = int(re.findall("V(\d+)", source_fp)[0])
            vol_no = "PageV{:02d}P{}".format(vol_no, "{:03d}")
        except:
            vol_no = "PageV01P{:03d}"
        def fmt_match(match):
            r = match.group(1) + vol_no.format(int(match.group(2)))
            return r + match.group(3)
        end = '<td class="book-page-show">|\Z'
        text = re.sub(r'(</td>)[^<\d]*(\d+)[^<]*({})'.format(end),
                      fmt_match, text)
        return text

    def remove_notes(self, text):
        """Remove footnotes from text, and format them as endnotes

        Footnotes are indicated in different ways in eShia texts;
        the only communality seems to be that a <hr/> tag
        splits off the critical apparatus from the text.
        We will use this feature to split off the text from the footnotes.
        """
        split_text = re.split("(PageV\d+P\d+)", text)
        text = []
        footnotes = []
        for i, t in enumerate(split_text):
            if re.match("PageV\d+P\d+", t):
                text.append(t)
            else: # check if the horizontal line splitting apparatus off is there
                spl = re.split("<hr */? *>", t)
                if len(spl) == 1: # no footnotes
                    text.append(t)
                else:
                    try:
                        notes = "Notes to {}:<br/>{}".format(split_text[i+1], spl[-1])
                    except:
                        notes = "Notes to [NO PAGE]:<br/>{}".format(spl[-1])
                    notes = html2md_eShia.markdownify(notes) 
                    footnotes.append(notes)
                    text.append("\n\n".join(spl[:-1]))

        text = "\n\n".join(text)
        notes = "\n\n".join(footnotes)
        notes = re.sub("\n+#* *\n+", "\n\n", notes)
        notes = self.endnote_splitter + notes
        return text, notes 
                    
    #def convert_html2md(self, html):
    def add_structural_annotations(self, html):
        """Convert html to mARkdown text using a html2md converter."""
        text = html2md_eShia.markdownify(html)
        return text             

    def post_process(self, text):
        empty = """ *اين صفحه در كتاب اصلي بدون متن است / هذه الصفحة فارغة في النسخة المطبوعة *"""
        text = re.sub(empty, "\n", text)
        text = super().post_process(text)
        return text

if __name__ == "__main__":
    import doctest
    doctest.testmod()
    input("Passed all tests. Continue?")
    
    conv = EShiaHtmlConverter()
    folder = r"G:\London\OpenITI\new\eShia"
    import os
    conv.convert_file(os.path.join(folder, "10461.html"))

    conv.convert_files_in_folder(folder)



"""Moves all text files from all subdirectories into a single directory.

Todo:
    if more than one version of the text file is in the folder,
        the script will not automatically copy the best version! 

Examples:
    Command line usage::
    
        $ collect_txt_files D:\\OpenITI\\25Y_folders D:\\OpenITI\\release\\texts

    If the source and target directories were not given,
    the user will be prompted for them::

        $ collect_txt_files
        Enter the source directory:
        Enter the target directory:
"""

import os
import re
import shutil
import sys

if __name__ == '__main__':
    from os import sys, path
    root_folder = path.dirname(path.dirname(path.dirname(path.abspath(__file__))))
    sys.path.append(root_folder)

from openiti.helper.uri import URI

def move_ara_files(source_dir, dest_dir):
    """ old name of the collect_text_files function;
    renamed because the name does not reflect what function does;
    old name kept for backward compatibility
    """
    return collect_text_files(source_dir, dest_dir)


def collect_text_files(source_dir, dest_dir,
                       exclude_folders=[],
                       extensions=["mARkdown", "inProgress", "completed", ""],
                       version_id_only=True):
    """Copy all text files in `source_dir` (+ its subfolders) to `dest_dir`.

    Args:
        source_dir (str): path to the directory that contains
            (subfolders containing) the text files
        dest_dir (str): path to a directory to which text files will be copied.
        exclude_folders (list): directories to be excluded
            from the collection process
        extensions (list): list of extensions; only text files with these
            extensions will be copied.
            Defaults to ["mARkdown", "inProgress", "completed", ""]
        version_id_only (bool): if True, the filename will be shortened
            to the last part of the URI, i.e., the version id + language id
            (e.g., Shamela001185-ara1). Defaults to True.
        
    Returns:
        (int): number of files copied
    """
    cnt = 0
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)
    for root, dirs, files in os.walk(source_dir):
        dirs[:] = [d for d in dirs if d not in exclude_folders]
        for file in files:
            #if re.search("^\d{4}\w+\.\w+\.\w+-\w{4}(\.(mARkdown|inProgress|completed))+$", file):
            # if re.search("^\d{4}\w+\.\w+\.\w+-ara\d$", file):
            try:
                uri = URI(os.path.join(root, file))
            except:
                uri = URI()
            if uri.uri_type == "version" and uri.extension in extensions:
                # to rename the files and keep only the text ID
                old_fp = os.path.join(root, file)
                if version_id_only:
                    new_fp = os.path.join(dest_dir, uri(ext="").rsplit(".")[-1])
                else:
                    new_fp = os.path.join(dest_dir, file)
                shutil.copy(old_fp, new_fp)
                #print(old_fp)
                #print(">", new_fp)
                cnt += 1
    return cnt


if __name__ == '__main__':

    if len(sys.argv) == 3:
        source = sys.argv[1]
        target = sys.argv[2]
    else:
        source = input("Enter the source directory: ")
        while not source:
            source = input("Enter the source directory: ")

        target = input("Enter the target directory: ")
        while not target:
            target = input("Enter the target directory: ")

    if not os.path.exists(source):
        print("Aborting. Source directory doesn't exist!")
        sys.exit()    
    cnt = move_ara_files(source, target)
    print("done; {} files copied".format(cnt))

    source = r"D:\London\OpenITI\25Y_repos"
    target = r"D:\London\OpenITI\python_library\openiti\openiti\test\0025AH"
    cnt = move_ara_files(source, target)
    print("done; {} files copied".format(cnt))
    sys.exit()



import re, os


#import zfunc
#import helper as hlp
from openiti.helper import funcs


def mechanical_chunking(path_full, target_folder):
    file_name = path_full.split("/")[-1]
    print("\n" + file_name)
    file_id = re.sub("\d{4}\w+\.\w+\.", "", file_name)
    print("\n" + file_id)

    target_path = target_folder + file_id

    # run generation
    cex = []

    if os.path.exists(target_path + "-%05d" % thresh):
        count = 0
    else:
        with open(path_full, "r", encoding="utf8") as f1:
            data = f1.read()
            #data = data.split(hlp.funcs.splitter)[1]
            data = data.split(funcs.splitter)[1]

            data = data.replace("\n~~", " ")
            data = data.replace("#", "(@@)")
            data = data.replace("\n", "(@)")
            data = re.sub(" +", " ", data)

            #data = data.split(hlp.funcs.milestone)
            data = data.split(funcs.milestone)

            print("\t%d items will be created..." % len(data))

            rec = '{"id":"%s", "series":"%s", "text": "%s"}'

            counter = 0
            for d in data:
                print("d: ", d)
                counter += 1
                id = file_id + ".ms%d" % counter
                #text = hlp.text_cleaner(d)
                text = funcs.text_cleaner(d)

                rec_t = rec % (id, file_id, text)

                cex.append(rec_t)

                #if counter % hlp.funcs.thresh == 0:
                if counter % funcs.thresh == 0:
                    with open(target_path + "-%05d" % counter, "w", encoding="utf8") as ft:
                        ft.write("\n".join(cex))
                    cex = []

            #counter_final = hlp.roundup(counter, hlp.funcs.thresh)
            counter_final = funcs.roundup(counter, funcs.thresh)
            with open(target_path + "-%05d" % counter_final, "w", encoding="utf8") as ft:
                ft.write("\n".join(cex))

            count = 1

    return count


# process all texts in OpenITI
def process_all(folder, target_path):

    print()
    print("Generating mechanical passim_new corpus...")
    print()

    count = 1

    for root, dirs, files in os.walk(folder):
        #dirs[:] = [d for d in dirs if d not in zfunc.exclude]
        dirs[:] = [d for d in dirs if d not in funcs.exclude_folders]

        for file in files:
            if re.search("^\d{4}\w+\.\w+\.\w+-\w{4}(\.(mARkdown|inProgress|completed))?$", file):
                path_full = os.path.join(root, file)
                # input(file)

                if "" in path_full:

                    # print(d)
                    count += mechanical_chunking(path_full, target_path)
                    return
                    if count % 100 == 0:
                        print()
                        print("=============" * 2)
                        print("Processed: %d" % count)
                        print("=============" * 2)
                        print()
                if count == 100:
                    break
                else:
                    continue



"""
Test relative speed of compiled regexes vs. normal regex

Result: no significant speed difference

Printout:

----------------------------------------
Testing functions re.sub(uncompiled_noise, '', text),
                  compiled_noise.sub('', text),
                  re.sub(compiled_noise, '', text),
                  re.sub(uncompiled_noise, '', text) second run:

All functions have the same outcome:  True
Results after 100 loops, with 30 first lines of the Qur'an:
1 - compiled_noise.sub('', text): 0.025620200000000003
2 - re.sub(compiled_noise, '', text): 0.027312799999999998
3 - re.sub(uncompiled_noise, '', text): 0.0312185
4 - re.sub(uncompiled_noise, '', text) second run: 0.0332567
(compiled_noise.sub('', text) ca. 1.1 times faster than re.sub(compiled_noise, '', text))
(compiled_noise.sub('', text) ca. 1.2 times faster than re.sub(uncompiled_noise, '', text))
(compiled_noise.sub('', text) ca. 1.3 times faster than re.sub(uncompiled_noise, '', text) second run)

Results after 10000 loops, with 30 first lines of the Qur'an:
1 - compiled_noise.sub('', text): 2.5880406999999996
2 - re.sub(compiled_noise, '', text): 2.5930370000000007
3 - re.sub(uncompiled_noise, '', text): 2.6693532
4 - re.sub(uncompiled_noise, '', text) second run: 2.7063191
(compiled_noise.sub('', text) ca. 1.0 times faster than re.sub(compiled_noise, '', text))
(compiled_noise.sub('', text) ca. 1.0 times faster than re.sub(uncompiled_noise, '', text))
(compiled_noise.sub('', text) ca. 1.0 times faster than re.sub(uncompiled_noise, '', text) second run)


Results after 100000 loops, with 30 first lines of the Qur'an:
1 - compiled_noise.sub('', text): 26.794016499999998
2 - re.sub(uncompiled_noise, '', text) second run: 26.809392700000004
3 - re.sub(uncompiled_noise, '', text): 27.1023572
4 - re.sub(compiled_noise, '', text): 27.329767200000006
(compiled_noise.sub('', text) ca. 1.0 times faster than re.sub(uncompiled_noise, '', text) second run)
(compiled_noise.sub('', text) ca. 1.0 times faster than re.sub(uncompiled_noise, '', text))
(compiled_noise.sub('', text) ca. 1.0 times faster than re.sub(compiled_noise, '', text))

----------------------------------------
Testing functions compiled_noise.sub('', text), re.sub(compiled_noise, '', text), re.sub(uncompiled_noise, '', text):

All functions have the same outcome:  True
Results after 10 loops, with large text of 75MB:
1 - re.sub(uncompiled_noise, '', text): 3.2946443999999957
2 - compiled_noise.sub('', text): 3.3465494000000007
3 - re.sub(compiled_noise, '', text): 3.352638099999993
(re.sub(uncompiled_noise, '', text) ca. 1.0 times faster than compiled_noise.sub('', text))
(re.sub(uncompiled_noise, '', text) ca. 1.0 times faster than re.sub(compiled_noise, '', text))


Results after 50 loops, with large text of 75MB:
1 - re.sub(compiled_noise, '', text): 16.902413100000018
2 - re.sub(uncompiled_noise, '', text): 16.9217314
3 - compiled_noise.sub('', text): 17.0021323
(re.sub(compiled_noise, '', text) ca. 1.0 times faster than re.sub(uncompiled_noise, '', text))
(re.sub(compiled_noise, '', text) ca. 1.0 times faster than compiled_noise.sub('', text))
"""

import re
import timeit

noiseC = re.compile(""" ّ    | # Tashdīd / Shadda
                       َ    | # Fatḥa
                       ً    | # Tanwīn Fatḥ / Fatḥatān
                       ُ    | # Ḍamma
                       ٌ    | # Tanwīn Ḍamm / Ḍammatān
                       ِ    | # Kasra
                       ٍ    | # Tanwīn Kasr / Kasratān
                       ْ    | # Sukūn
                       ۡ    | # Quranic Sukūn
                       ࣰ    | # Quranic Open Fatḥatān
                       ࣱ    | # Quranic Open Ḍammatān
                       ࣲ    | # Quranic Open Kasratān
                       ٰ    | # Dagger Alif
                       ـ     # Taṭwīl / Kashīda
                   """, re.VERBOSE)
noise = '[ًࣰٌࣱٍࣲَُِّْٰۡـ]'



def time_functions(functions, loops, string_descr):
    """Check which function is faster.

    Args:
        functions (list): a list of tuples (function_name (str),
                                            function_descr (str))
        loops (list): a list of integers, representing the numbers
            of times a pair of functions should be timed.
        string_descr (str): a description of the test string.
    NB: the functions to be tested should not have arguments.
    """
    print("--"*20)
    print("Testing functions {}:".format(", ".join([y for x,y in functions])))
    print()
    accuracy = (len(set([eval(f[0]+'()') for f in functions])) == 1)
    print("All functions have the same outcome: ", accuracy)
    for n in loops:
        results = []
        for f, f_descr in functions:
            t = timeit.timeit(f+"()", setup="from __main__ import "+f, number=n)
            results.append([f_descr, t])
        results = sorted(results, key=lambda item: item[1])
        print("Results after {} loops, with {}:".format(n, string_descr))
        for i, x in enumerate(results):
            print("{} - {}: {}".format(i+1, x[0], x[1]))
        for x in results[1:]:
            fmt = "({} ca. {} times faster than {})"
            print(fmt.format(results[0][0], round(x[1]/results[0][1], 1), x[0]))
        print()


def re_sub_deNoise_test1():
    return noiseC.sub('', text)

def re_sub_deNoise_test2():
    return re.sub(noiseC, '', text)

def re_sub_deNoise_test3():
    return re.sub(noise, '', text)

page = """
بِسْمِ اللَّهِ الرَّحْمَنِ الرَّحِيمِ
الْحَمْدُ لِلَّهِ رَبِّ الْعَالَمِينَ
الرَّحْمَنِ الرَّحِيمِ
مَالِكِ يَوْمِ الدِّينِ
إِيَّاكَ نَعْبُدُ وَإِيَّاكَ نَسْتَعِينُ
اهْدِنَا الصِّرَاطَ الْمُسْتَقِيمَ
صِرَاطَ الَّذِينَ أَنْعَمْتَ عَلَيْهِمْ غَيْرِ الْمَغْضُوبِ عَلَيْهِمْ وَلَا الضَّالِّينَ
بِسْمِ اللَّهِ الرَّحْمَنِ الرَّحِيمِ الم
ذَلِكَ الْكِتَابُ لَا رَيْبَ فِيهِ هُدًى لِلْمُتَّقِينَ
الَّذِينَ يُؤْمِنُونَ بِالْغَيْبِ وَيُقِيمُونَ الصَّلَاةَ وَمِمَّا رَزَقْنَاهُمْ يُنْفِقُونَ
وَالَّذِينَ يُؤْمِنُونَ بِمَا أُنْزِلَ إِلَيْكَ وَمَا أُنْزِلَ مِنْ قَبْلِكَ وَبِالْآخِرَةِ هُمْ يُوقِنُونَ
أُولَئِكَ عَلَى هُدًى مِنْ رَبِّهِمْ وَأُولَئِكَ هُمُ الْمُفْلِحُونَ
إِنَّ الَّذِينَ كَفَرُوا سَوَاءٌ عَلَيْهِمْ أَأَنْذَرْتَهُمْ أَمْ لَمْ تُنْذِرْهُمْ لَا يُؤْمِنُونَ
خَتَمَ اللَّهُ عَلَى قُلُوبِهِمْ وَعَلَى سَمْعِهِمْ وَعَلَى أَبْصَارِهِمْ غِشَاوَةٌ وَلَهُمْ عَذَابٌ عَظِيمٌ
وَمِنَ النَّاسِ مَنْ يَقُولُ آمَنَّا بِاللَّهِ وَبِالْيَوْمِ الْآخِرِ وَمَا هُمْ بِمُؤْمِنِينَ
يُخَادِعُونَ اللَّهَ وَالَّذِينَ آمَنُوا وَمَا يَخْدَعُونَ إِلَّا أَنْفُسَهُمْ وَمَا يَشْعُرُونَ
فِي قُلُوبِهِمْ مَرَضٌ فَزَادَهُمُ اللَّهُ مَرَضًا وَلَهُمْ عَذَابٌ أَلِيمٌ بِمَا كَانُوا يَكْذِبُونَ
وَإِذَا قِيلَ لَهُمْ لَا تُفْسِدُوا فِي الْأَرْضِ قَالُوا إِنَّمَا نَحْنُ مُصْلِحُونَ
أَلَا إِنَّهُمْ هُمُ الْمُفْسِدُونَ وَلَكِنْ لَا يَشْعُرُونَ
وَإِذَا قِيلَ لَهُمْ آمِنُوا كَمَا آمَنَ النَّاسُ قَالُوا أَنُؤْمِنُ كَمَا آمَنَ السُّفَهَاءُ أَلَا إِنَّهُمْ هُمُ السُّفَهَاءُ وَلَكِنْ لَا يَعْلَمُونَ
وَإِذَا لَقُوا الَّذِينَ آمَنُوا قَالُوا آمَنَّا وَإِذَا خَلَوْا إِلَى شَيَاطِينِهِمْ قَالُوا إِنَّا مَعَكُمْ إِنَّمَا نَحْنُ مُسْتَهْزِئُونَ
اللَّهُ يَسْتَهْزِئُ بِهِمْ وَيَمُدُّهُمْ فِي طُغْيَانِهِمْ يَعْمَهُونَ
أُولَئِكَ الَّذِينَ اشْتَرَوُا الضَّلَالَةَ بِالْهُدَى فَمَا رَبِحَتْ تِجَارَتُهُمْ وَمَا كَانُوا مُهْتَدِينَ
مَثَلُهُمْ كَمَثَلِ الَّذِي اسْتَوْقَدَ نَارًا فَلَمَّا أَضَاءَتْ مَا حَوْلَهُ ذَهَبَ اللَّهُ بِنُورِهِمْ وَتَرَكَهُمْ فِي ظُلُمَاتٍ لَا يُبْصِرُونَ
صُمٌّ بُكْمٌ عُمْيٌ فَهُمْ لَا يَرْجِعُونَ
أَوْ كَصَيِّبٍ مِنَ السَّمَاءِ فِيهِ ظُلُمَاتٌ وَرَعْدٌ وَبَرْقٌ يَجْعَلُونَ أَصَابِعَهُمْ فِي آذَانِهِمْ مِنَ الصَّوَاعِقِ حَذَرَ الْمَوْتِ وَاللَّهُ مُحِيطٌ بِالْكَافِرِينَ
يَكَادُ الْبَرْقُ يَخْطَفُ أَبْصَارَهُمْ كُلَّمَا أَضَاءَ لَهُمْ مَشَوْا فِيهِ وَإِذَا أَظْلَمَ عَلَيْهِمْ قَامُوا وَلَوْ شَاءَ اللَّهُ لَذَهَبَ بِسَمْعِهِمْ وَأَبْصَارِهِمْ إِنَّ اللَّهَ عَلَى كُلِّ شَيْءٍ قَدِيرٌ
يَا أَيُّهَا النَّاسُ اعْبُدُوا رَبَّكُمُ الَّذِي خَلَقَكُمْ وَالَّذِينَ مِنْ قَبْلِكُمْ لَعَلَّكُمْ تَتَّقُونَ
الَّذِي جَعَلَ لَكُمُ الْأَرْضَ فِرَاشًا وَالسَّمَاءَ بِنَاءً وَأَنْزَلَ مِنَ السَّمَاءِ مَاءً فَأَخْرَجَ بِهِ مِنَ الثَّمَرَاتِ رِزْقًا لَكُمْ فَلَا تَجْعَلُوا لِلَّهِ أَنْدَادًا وَأَنْتُمْ تَعْلَمُونَ
وَإِنْ كُنْتُمْ فِي رَيْبٍ مِمَّا نَزَّلْنَا عَلَى عَبْدِنَا فَأْتُوا بِسُورَةٍ مِنْ مِثْلِهِ وَادْعُوا شُهَدَاءَكُمْ مِنْ دُونِ اللَّهِ إِنْ كُنْتُمْ صَادِقِينَ
"""

fp = r"D:\London\OpenITI\new_books\bok\ShamelaRARs\converted\4_OpenITI_mARkdown_new\Shamela026332.txt"
with open(fp, mode="r", encoding="utf-8") as file:
    large_text = file.read()

text = page  # reset the global text variable

time_functions([("re_sub_deNoise_test3", "re.sub(uncompiled_noise, '', text)"),
                ("re_sub_deNoise_test1", "compiled_noise.sub('', text)"),
                ("re_sub_deNoise_test2", "re.sub(compiled_noise, '', text)"),
                ("re_sub_deNoise_test3", "re.sub(uncompiled_noise, '', text) second run")],
               [100, 10000, 100000],
               "30 first lines of the Qur'an")

text = large_text  # reset the global text variable

time_functions([("re_sub_deNoise_test1", "compiled_noise.sub('', text)"),
                ("re_sub_deNoise_test2", "re.sub(compiled_noise, '', text)"),
                ("re_sub_deNoise_test3", "re.sub(uncompiled_noise, '', text)")],
                [10, 50],
               "large text of 75MB")




import re
import unicodedata
import urllib
import doctest

ar_chars = """\
ء	ARABIC LETTER HAMZA
آ	ARABIC LETTER ALEF WITH MADDA ABOVE
أ	ARABIC LETTER ALEF WITH HAMZA ABOVE
ؤ	ARABIC LETTER WAW WITH HAMZA ABOVE
إ	ARABIC LETTER ALEF WITH HAMZA BELOW
ئ	ARABIC LETTER YEH WITH HAMZA ABOVE
ا	ARABIC LETTER ALEF
ب	ARABIC LETTER BEH
ة	ARABIC LETTER TEH MARBUTA
ت	ARABIC LETTER TEH
ث	ARABIC LETTER THEH
ج	ARABIC LETTER JEEM
ح	ARABIC LETTER HAH
خ	ARABIC LETTER KHAH
د	ARABIC LETTER DAL
ذ	ARABIC LETTER THAL
ر	ARABIC LETTER REH
ز	ARABIC LETTER ZAIN
س	ARABIC LETTER SEEN
ش	ARABIC LETTER SHEEN
ص	ARABIC LETTER SAD
ض	ARABIC LETTER DAD
ط	ARABIC LETTER TAH
ظ	ARABIC LETTER ZAH
ع	ARABIC LETTER AIN
غ	ARABIC LETTER GHAIN
ـ	ARABIC TATWEEL
ف	ARABIC LETTER FEH
ق	ARABIC LETTER QAF
ك	ARABIC LETTER KAF
ل	ARABIC LETTER LAM
م	ARABIC LETTER MEEM
ن	ARABIC LETTER NOON
ه	ARABIC LETTER HEH
و	ARABIC LETTER WAW
ى	ARABIC LETTER ALEF MAKSURA
ي	ARABIC LETTER YEH
ً	ARABIC FATHATAN
ٌ	ARABIC DAMMATAN
ٍ	ARABIC KASRATAN
َ	ARABIC FATHA
ُ	ARABIC DAMMA
ِ	ARABIC KASRA
ّ	ARABIC SHADDA
ْ	ARABIC SUKUN
٠	ARABIC-INDIC DIGIT ZERO
١	ARABIC-INDIC DIGIT ONE
٢	ARABIC-INDIC DIGIT TWO
٣	ARABIC-INDIC DIGIT THREE
٤	ARABIC-INDIC DIGIT FOUR
٥	ARABIC-INDIC DIGIT FIVE
٦	ARABIC-INDIC DIGIT SIX
٧	ARABIC-INDIC DIGIT SEVEN
٨	ARABIC-INDIC DIGIT EIGHT
٩	ARABIC-INDIC DIGIT NINE
ٮ	ARABIC LETTER DOTLESS BEH
ٰ	ARABIC LETTER SUPERSCRIPT ALEF
ٹ	ARABIC LETTER TTEH
پ	ARABIC LETTER PEH
چ	ARABIC LETTER TCHEH
ژ	ARABIC LETTER JEH
ک	ARABIC LETTER KEHEH
گ	ARABIC LETTER GAF
ی	ARABIC LETTER FARSI YEH
ے	ARABIC LETTER YEH BARREE
۱	EXTENDED ARABIC-INDIC DIGIT ONE
۲	EXTENDED ARABIC-INDIC DIGIT TWO
۳	EXTENDED ARABIC-INDIC DIGIT THREE
۴	EXTENDED ARABIC-INDIC DIGIT FOUR
۵	EXTENDED ARABIC-INDIC DIGIT FIVE
۶	EXTENDED ARABIC-INDIC DIGIT SIX
۷	EXTENDED ARABIC-INDIC DIGIT SEVEN
۸	EXTENDED ARABIC-INDIC DIGIT EIGHT
۹	EXTENDED ARABIC-INDIC DIGIT NINE
۰	EXTENDED ARABIC-INDIC DIGIT ZERO
"""
##‌	ZERO WIDTH NON-JOINER
##‍	ZERO WIDTH JOINER"""
ar_chars = [x.split("\t")[0] for x in ar_chars.splitlines()]
#ar_chars = "ذ١٢٣٤٥٦٧٨٩٠ّـضصثقفغعهخحجدًٌَُلإإشسيبلاتنمكطٍِلأأـئءؤرلاىةوزظْلآآ"
ar_char = re.compile("[{}]".format("".join(ar_chars))) # regex for one Arabic character
ar_tok = re.compile("[{}]+".format("".join(ar_chars))) # regex for one Arabic token
noise = re.compile(""" ّ    | # Tashdīd / Shadda
                       َ    | # Fatḥa
                       ً    | # Tanwīn Fatḥ / Fatḥatān
                       ُ    | # Ḍamma
                       ٌ    | # Tanwīn Ḍamm / Ḍammatān
                       ِ    | # Kasra
                       ٍ    | # Tanwīn Kasr / Kasratān
                       ْ    | # Sukūn
                       ۡ    | # Quranic Sukūn
                       ࣰ    | # Quranic Open Fatḥatān
                       ࣱ    | # Quranic Open Ḍammatān
                       ࣲ    | # Quranic Open Kasratān
                       ٰ    | # Dagger Alif
                       ـ     # Taṭwīl / Kashīda
                   """, re.VERBOSE)

def denoise(text):
    """Remove non-consonantal characters from Arabic text.

    Examples:
        >>> denoise("وَالَّذِينَ يُؤْمِنُونَ بِمَا أُنْزِلَ إِلَيْكَ وَمَا أُنْزِلَ مِنْ قَبْلِكَ وَبِالْآخِرَةِ هُمْ يُوقِنُونَ")
        'والذين يؤمنون بما أنزل إليك وما أنزل من قبلك وبالآخرة هم يوقنون'
        >>> denoise(" ْ ً ٌ ٍ َ ُ ِ ّ ۡ ࣰ ࣱ ࣲ ٰ ")
        '              '
    """
    return re.sub(noise, "", text)


deNoise = denoise


def normalize(text, replacement_tuples=[]):
    """Normalize Arabic text by replacing complex characters by simple ones.
    The function is used internally to do batch replacements. Also, it can be called externally
     to run custom replacements with a list of tuples of (character/regex, replacement).

    Args:
        text (str): the string that needs to be normalized
        replacement_tuples (list of tuple pairs): (character/regex, replacement)

    Examples:
        >>> normalize('AlphaBet', [("A", "a"), ("B", "b")])
        'alphabet'
    """
    for pat, repl in replacement_tuples:
        text = re.sub(pat, repl, text)
    return text


def normalize_per(text):
    """Normalize Persian strings by converting Arabic chars to related Persian unicode chars.
    fixing Alifs, Alif Maqsuras, hamzas, ta marbutas, kaf, ya، Fathatan, kasra;

    Args:
        text (str): user input string to be normalized

    Examples:
        >>> normalize_per("سياسي")
        'سیاسی'
        >>> normalize_per("مدينة")
        'مدینه'
        >>> normalize_per("درِ باز")
        'در باز'
        >>> normalize_per("حتماً")
        'حتما'
        >>> normalize_per("مدرك")
        'مدرک'
        >>> normalize_per("أسماء")
        'اسما'
        >>> normalize_per("دربارۀ")
        'درباره'

    """

    repl = [
        ('ك', 'ک'),
        ('[أاإٱ]', 'ا'),
        ('[يى]ء?', 'ی'),
        ('ؤِ', 'و'),
        ('ئ', 'ی'),
        ('[ءًِ]', ''),
        ('[ۀة]', 'ه')
    ]

    return normalize(text, repl)


def normalize_ara_light(text):
    """Lighlty normalize Arabic strings:
    fixing only Alifs, Alif Maqsuras;
    replacing hamzas on carriers with standalone hamzas

    Args:
        text (str): the string that needs to be normalized

    Examples:
        >>> normalize_ara_light("ألف الف إلف آلف ٱلف")
        'الف الف الف الف الف'
        >>> normalize_ara_light("يحيى")
        'يحيي'
        >>> normalize_ara_light("مقرئ فيء")
        'مقرء فء'
        >>> normalize_ara_light("قهوة")
        'قهوة'
    """
    text = normalize_composites(text)
    repl = [("أ", "ا"), ("ٱ", "ا"), ("آ", "ا"), ("إ", "ا"),    # alifs
            ("ى", "ي"),                                        # alif maqsura
            ("يء", "ء"), ("ىء", "ء"), ("ؤ", "ء"), ("ئ", "ء"),  # hamzas
            ]
    return normalize(text, repl)
    

def normalize_ara_heavy(text):
    """Normalize Arabic text by simplifying complex characters:
    alifs, alif maqsura, hamzas, ta marbutas

    Examples:
        >>> normalize_ara_heavy("ألف الف إلف آلف ٱلف")
        'الف الف الف الف الف'
        >>> normalize_ara_heavy("يحيى")
        'يحيي'
        >>> normalize_ara_heavy("مقرئ فيء")
        'مقر في'
        >>> normalize_ara_heavy("قهوة")
        'قهوه'
    """
    text = normalize_composites(text)
    repl = [("أ", "ا"), ("ٱ", "ا"), ("آ", "ا"), ("إ", "ا"),  # alifs
            ("ى", "ي"),                                      # alif maqsura
            ("ؤ", ""), ("ئ", ""), ("ء", ""),                 # hamzas
            ("ة", "ه")                                       # ta marbuta
            ]
    return normalize(text, repl)


def normalize_composites(text, method="NFKC"):
    """Normalize composite characters and ligatures

    Ligatures like "Allah" will be broken into their components,
    while combining characters like alif + hamza above will be joined
    into one character. 

    Args:
        text (str): the string to be normalized
        method (str): the unicode method to be used for normalization
            (see https://docs.python.org/3.5/library/unicodedata.html).
            Default: NFKC, which is most suited for Arabic. 

    Examples:
        >>> len("ﷲ") # U+FDF2: ARABIC LIGATURE ALLAH ISOLATED FORM
        1
        >>> len(normalize_composites("ﷲ"))
        4
        >>> [char for char in normalize_composites("ﷲ")]
        ['ا', 'ل', 'ل', 'ه']
        
        >>> len("ﻹ") # UFEF9: ARABIC LIGATURE LAM WITH ALEF WITH HAMZA BELOW ISOLATED FORM
        1
        >>> len(normalize_composites("ﻹ"))
        2

        alif+hamza written with 2 unicode characters:
        U+0627 (ARABIC LETTER ALEF) + U+0654 (ARABIC HAMZA ABOVE):
        
        >>> a = "أ"
        >>> len(a)
        2
        >>> len(normalize_composites(a))
        1
    """
    return unicodedata.normalize(method, text)


def denormalize(text):
    """Replace complex characters with a regex covering all variants.

    Examples:
        >>> denormalize("يحيى")
        'يحي[يى]'
        >>> denormalize("هوية")
        'هوي[هة]'
        >>> denormalize("مقرئ")
        'مقر(?:[ؤئ]|[وي]ء)'
        >>> denormalize("فيء")
        'في(?:[ؤئ]|[وي]ء)'
    """
    alifs = '[إأٱآا]'
    alif_maqsura = '[يى]\\b'
    alif_maqsura_reg = '[يى]'
    ta_marbuta = '[هة]\\b'
    ta_marbuta_reg = '[هة]'
    hamzas = '[ؤئء]'
    hamzas_reg = '(?:{}|{}ء)'.format('[ؤئ]', '[وي]')
    # Applying deNormalization
    text = re.sub(alifs, alifs, text)
    text = re.sub(alif_maqsura, alif_maqsura_reg, text)
    text = re.sub(ta_marbuta, ta_marbuta_reg, text)
    text = re.sub(hamzas, hamzas_reg, text)
    return text


#def ar_ch_len(fp):
def ar_cnt_file(fp, mode="token", incl_editor_sections=True):
    """Count the number of Arabic characters/tokens in a text, given its pth

    Args:
        fp (str): url / path to a file
        mode (str): either "char" for count of Arabic characters,
                    or "token" for count of Arabic tokens
        incl_editor_sections (bool): if False, the sections marked as editorial
            (### |EDITOR|) will be left out of the token/character count.
            Default: True (editorial sections will be counted)

    Returns:
        (int): Arabic character/token count 
    """
    splitter = "#META#Header#End#"
    try:
        with urllib.request.urlopen(fp) as f:
            book = f.read().decode('utf-8')
    except:
        with open(fp, mode="r", encoding="utf-8") as f:
            book = f.read()

    if splitter in book:
        text = book.split(splitter)[-1]
    else:
        text = book
        msg = "This text is missing the splitter!\n{}".format(fp)
        #raise Exception(msg)
    if not incl_editor_sections:
        text = re.sub(r"### \|EDITOR.+?(### |\Z)", r"\1", text,
                      flags = re.DOTALL)

    # count the number of Arabic letters or tokens:
    
    if mode == "char":
        return ar_ch_cnt(text)
    else:
        return ar_tok_cnt(text)



def ar_ch_cnt(text):
    """
    Count the number of Arabic characters in a string

    :param text: text
    :return: number of the Arabic characters in the text

    Examples:
        >>> a = "ابجد ابجد اَبًجٌدُ"
        >>> ar_ch_cnt(a)
        16
    """
    return len(ar_char.findall(text))


def ar_tok_cnt(text):
    """
    Count the number of Arabic tokens in a string

    :param text: text
    :return: number of Arabic tokens in the text

    Examples:
        >>> a = "ابجد ابجد اَبًجٌدُ"
        >>> ar_tok_cnt(a)
        3
    """
    return len(ar_tok.findall(text))


def tokenize(text, token_regex=ar_tok):
    """Tokenize a text into tokens defined by `token_regex`

    NB: make sure to remove the OpenITI header from the text

    Args:
        text (str): full text with OpenITI header removed,
            cleaned of order marks ("\u202a", "\u202b", "\u202c")
        token_regex (str): regex that defines a token
        
    Returns:
        tuple (tokens (list): list of all tokens in the text,
               tokenStarts (list): list of start index of each token
               tokenEnds (list): list of end index of each token
               )
    Examples:
        >>> a = "ابجد ابجد اَبًجٌدُ"
        >>> tokens, tokenStarts, tokenEnds = tokenize(a)
        >>> tokens
        ['ابجد', 'ابجد', 'اَبًجٌدُ']
        >>> tokenStarts
        [0, 5, 10]
        >>> tokenEnds
        [4, 9, 18]
    """
    #find matches
    tokens = [m for m in re.finditer(token_regex,text)]

    #extract tokens and start,end positions
    tokenStarts = [m.start() for m in tokens]
    tokenEnds = [m.end() for m in tokens]
    tokens = [m.group() for m in tokens]

    return tokens, tokenStarts, tokenEnds

if __name__ == "__main__":
    doctest.testmod()



"""Scripts to add books to the correct OpenITI repositories.

* initialize_texts_from_csv: use a csv file that contains for every file
    that needs to be added to the corpus:
    - the path to its current location
    - the URI it needs to get

* initialize_new_texts_in_folder: initialize all text files in a folder
    (in order for this to work, all files need to have valid URIs)

* download_texts_from_CSV: download texts from the internet and
    directly put them into the correct OpenITI folder


"""

import copy
import os
import re
import requests
import shutil

if __name__ == '__main__':
    from os import sys, path
    root_folder =path.dirname(path.dirname(path.abspath(__file__)))
    root_folder = path.dirname(path.dirname(root_folder))
    sys.path.append(root_folder)

from openiti.helper.funcs import read_header
from openiti.helper.ara import ar_cnt_file
from openiti.helper.uri import move_to_new_uri_pth, add_character_count, URI, new_yml


created_ymls = []

def initialize_texts_from_CSV(csv_fp, old_base_pth="", new_base_pth="",
                              execute=False):
    """
    Use a CSV file (filename, URI) to move a list of texts to the relevant \
    OpenITI folder.

    The CSV file (which should not contain a heading) can contain
    full filepaths to the original files, or only filenames;
    in the latter case, the path to the folder where these files are located
    should be passed to the function as the old_base_pth argument.
    Similarly, the URI column can contain full OpenITI URI filepaths
    or only the URIs; in the latter case, the path to the folder
    containing the OpenITI 25-years folders should be passed to the function
    as the new_base_pth argument.

    Args:
        csv_fp (str): path to a csv file that contains the following columns:
            0. filepath to (or filename of) the text file
            1. full version uri of the text file
            (no headings!)
        old_base_path (str): path to the folder containing
            the files that need to be initialized
        new_base_pth (str): path to the folder containing
            the OpenITI 25-years repos
        execute (bool): if False, the proposed changes will only be printed
            (the user will still be given the option to execute
            all proposed changes at the end);
            if True, all changes will be executed immediately.

    Returns:
        None
    """
    with open(csv_fp, mode="r", encoding="utf-8") as file:
        csv = file.read().splitlines()
        csv = [re.split("[,\t]", row) for row in csv]

    for old_fp, new in csv:
        if old_base_pth:
            old_fp = os.path.join(old_base_pth, old_fp)
        new_uri = URI(new)
        if new_base_pth:
            new_uri.base_pth = new_base_pth
        #char_count = ar_ch_len(old_fp)
        tok_count = ar_cnt_file(old_fp, mode="token")
        char_count = ar_cnt_file(old_fp, mode="char")

        move_to_new_uri_pth(old_fp, new_uri, execute)

        add_character_count(tok_count, char_count, new_uri, execute)

    if not execute:
        resp = input("To carry out these changes: press OK+Enter; \
to abort: press Enter. ")
        if resp == "OK":
            initialize_texts_from_CSV(csv_fp, old_base_pth, new_base_pth,
                              execute=True)
        else:
            print()
            print("User aborted carrying out these changes!")
            print("*"*60)


def initialize_new_texts_in_folder(folder, target_base_pth, execute=False):
    """Move all new texts in folder to their OpenITI repo, creating yml files\
    if necessary (or copying them from the same folder if present).

    Args:
        folder (str): path to the folder that contains new text files
            (with OpenITI uri filenames) and perhaps yml files
        target_base_pth (str): path to the folder containing the 25-years repos
        execute (bool): if False, the proposed changes will only be printed
            (the user will still be given the option to execute
            all proposed changes at the end);
            if True, all changes will be executed immediately.

    Returns:
        None

    Examples::
    
        # >>> folder = r"D:\\OpenITI\\barzakh"
        # >>> target_base_pth = r"D:\\OpenITI\\25Yrepos"
        # >>> initialize_new_texts_in_folder(folder, target_base_pth,
        #                                    execute=False)
    """
    for fn in os.listdir(folder):
        ext = os.path.splitext(fn)[1]
        if ext not in (".yml", ".md"):
            fp = os.path.join(folder, fn)
            print(fp)
            if os.path.isfile(fp):
                initialize_new_text(fp, target_base_pth, execute)


def initialize_new_text(origin_fp, target_base_pth, execute=False):
    """Move a new text file to its OpenITI repo, creating yml files\
    if necessary (or copying them from the same folder if present).

    The function also checks whether the new text adheres to OpenITI text format.

    Args:
        origin_fp (str): filepath of the text file (filename must be
                         in OpenITI uri format)
        target_base_pth (str): path to the folder
                               that contains the 25-years-repos
        execute (bool): if False, the proposed changes will only be printed
            (the user will still be given the option to execute
            all proposed changes at the end);
            if True, all changes will be executed immediately.

    Returns:
        None

    Example:
        # >>> origin_folder = r"D:\\OpenITI\\barzakh"
        # >>> fn = "0375IkhwanSafa.Rasail.Hindawi95926405Vols-ara1.completed"
        # >>> origin_fp = os.path.join(origin_folder, fn)
        # >>> target_base_pth = r"D:\\OpenITI\\25Yrepos"
        # >>> initialize_new_text(origin_fp, target_base_pth, execute=False)
    """
    ori_uri = URI(origin_fp)
    tar_uri = copy.deepcopy(ori_uri)
    tar_uri.base_pth = target_base_pth
    target_fp = tar_uri.build_pth("version_file")

    # Check whether the text file has OpenITI format:

    header = "\n".join(read_header(origin_fp))
    if "#META#Header#End" not in header:
        print("Initialization aborted: ")
        print("{} does not contain OpenITI metadata header splitter!".format(origin_fp))
        return
    if "######OpenITI#" not in header:
        print("Initialization aborted: ")
        print("{} does not contain OpenITI magic value!".format(origin_fp))
        return

    # Count the Arabic characters in the text file:

    #tok_count = ar_ch_len(origin_fp)
    tok_count = ar_cnt_file(origin_fp, mode="token")
    char_count = ar_cnt_file(origin_fp, mode="char")

    # Move the text file:

    target_folder = tar_uri.build_pth("version")
    move_to_new_uri_pth(origin_fp, tar_uri, execute)

    # Move or create the YML files:

    for yf in ("version_yml", "book_yml", "author_yml"):
        yfp = os.path.join(ori_uri.base_pth, ori_uri.build_uri(yf))
        tar_yfp = tar_uri.build_pth(yf)
        if os.path.exists(yfp):
            if execute:
                shutil.move(yfp, tar_yfp)
            else:
                print("  move", yfp, "to", tar_yfp)
        else:
            if not os.path.exists(tar_yfp):
                new_yml(tar_yfp, yf, execute)
            else:
                if tar_yfp not in created_ymls:
                    if not execute:
                        print("  {} already exist; no yml file created".format(tar_yfp))


    # Add the character count to the new yml file:

    add_character_count(tok_count, char_count, tar_uri, execute)

    # Give the option to execute the changes:

    if not execute:
        print("Execute these changes?")
        resp = input("Type OK + Enter to execute; press Enter to abort: ")
        if resp == "OK":
            initialize_new_text(origin_fp, target_base_pth, execute=True)
        else:
            print("User aborted the execution of the changes.")


def download_texts_from_CSV(csv_fp, base_url="", new_base_pth=""):
    """
    Use a CSV file (filename, URI) to download a list of texts to the relevant \
    OpenITI folder.

    The CSV file (which should not contain a heading) can contain
    full urls for the original files, or only filenames;
    in the latter case, the url of the website where these files are located
    should be passed to the function as the base_url argument.
    Similarly, the URI column can contain full OpenITI URI filepaths
    or only the URIs; in the latter case, the path to the folder
    containing the OpenITI 25-years folders should be passed to the function
    as the new_base_pth argument.

    Args:
        csv_fp (str): path to a csv file that contains the following columns:
            0. filepath to (or filename of) the text file
            1. full version uri of the text file
            (no headings!)
        old_base_path (str): path to the folder containing
            the files that need to be initialized. Defaults to "".
        new_base_pth (str): path to the folder containing
            the OpenITI 25-years repos. Defaults to "".
    
    Returns:
        None
    """
    with open(csv_fp, mode="r", encoding="utf-8") as file:
        csv = file.read().splitlines()
        csv = [re.split("[,\t]", row) for row in csv]

    temp_folder = "temp"
    if not os.path.exists(temp_folder):
        os.makedirs(temp_folder)

    for old_fp, new in csv:
        print(old_fp)
        if not os.path.exists(new):
            if base_url:
                old_fp = os.path.join(base_url, old_fp)
            new_uri = URI(new)
            if new_base_pth:
                new_uri.base_pth = new_base_pth
            #char_count = ar_ch_len(old_fp)

            fn = os.path.split(old_fp)[1]
            temp_fp = os.path.join(temp_folder, fn)

            with requests.get(old_fp, stream=True) as r:
                r.raise_for_status()
                print("download starting")
                with open(temp_fp, mode="wb") as f:
                    for chunk in r.iter_content(chunk_size=8192):
                        if chunk:
                            f.write(chunk)
                print("download finished")
            

            move_to_new_uri_pth(temp_fp, new_uri, execute=True)

            if not temp_fp.endswith("pdf") and not temp_fp.endswith("zip"):
                tok_count = ar_cnt_file(temp_fp, mode="token")
                char_count = ar_cnt_file(temp_fp, mode="char")
                add_character_count(tok_count, char_count, new_uri, execute=True)

    shutil.rmtree(temp_folder)



if __name__ == "__main__":
    base_pth = r"D:\London\OpenITI\python_library\openiti\openiti\test"

##    # test initialize_new_texts_in_folder function:
##    barzakh = os.path.join(base_pth, "barzakh")
##    initialize_new_texts_in_folder(barzakh, base_pth, execute=True)
##    print("Texts in test/barzakh initialized; check if initialization was successful!")
##    input("Press Enter to continue")

##    # test initialize_texts_from_CSV function:
##    csv_fp = os.path.join(base_pth, "initialize.csv")
##    initialize_texts_from_CSV(csv_fp, old_base_pth="", new_base_pth=base_pth,
##                              execute=False)
##    print("Texts in csv file initialized; check if initialization was successful!")
##    input("Press Enter to continue")

##    # test download_texts_from_CSV function:
##    csv_fp = os.path.join(base_pth, "download_csv_to_uri.csv")
##    download_texts_from_CSV(csv_fp, new_base_pth=base_pth)
##    print("Texts in csv file downloaded and in the right folder. Check!")

    folder = r"D:\London\OpenITI\barzakh\barzakh"
    target_base_pth = r"D:\London\OpenITI\25Y_repos"
    ignore = (".yml", ".md", ".py", ".git", ".gitignore")
    for fn in os.listdir(folder):
        if not fn.endswith(ignore) and "oorlib" not in fn:
            fp = os.path.join(folder, fn)
            print(fp)
            if os.path.isfile(fp):
                initialize_new_text(fp, target_base_pth, execute=False)
